"""Calculate SOL momentum loss as a function target electron temperature.

See Figure 15 of :cite:`stangeby_2018`.
"""
import numpy as np

from ...named_options import MomentumLossFunction
from ...unit_handling import Quantity, ureg, wraps_ufunc


def _calc_SOL_momentum_loss_fraction(A: float, Tstar: float, n: float, target_electron_temp: float) -> float:
    """Calculates the fraction of momentum lost in the SOL, for a generic SOL momentum loss function.

    This is equation 33 of :cite:`stangeby_2018`, rearranged for $f^{total}_{mom-loss}$
    """
    return float(1.0 - A * (1.0 - np.exp(-target_electron_temp / Tstar)) ** n)


@wraps_ufunc(
    return_units=dict(momentum_loss_fraction=ureg.dimensionless),
    input_units=dict(key=None, target_electron_temp=ureg.eV),
)
def calc_SOL_momentum_loss_fraction(key: MomentumLossFunction, target_electron_temp: Quantity) -> float:
    """Calculate the fraction of momentum lost in the SOL.

    The coefficients come from figure captions in :cite:`stangeby_2018`
    * KotovReiter: SOLPS scans with Deuterium only, no impurities for JET vertical target. Figure 7a)
    * Sang: SOLPS scans for Deuterium with Carbon impurity, for DIII-D, variety of divertor configurations. Figure 7b)
    * Jarvinen: EDGE2D density scan for JET with a horizontal target, for a variety of targets and injected impurities. Figure 10a)
    * Moulton: SOLPS density scan for narrow slot divertor, no impurities. Figure 10b)
    * PerezH: SOLPS density scan for AUG H-mode, only trace impurities. Figure 11a)
    * PerezL: SOLPS density scan for AUG L-mode, with Carbon impurity. Figure 11b)

    Comparison is in Figure 15.

    Args:
        key: which momentum loss function to use
        target_electron_temp: electron temperature at the target [eV]

    Returns:
        SOL_momentum_loss_fraction [~]
    """
    if key == MomentumLossFunction.KotovReiter:
        return _calc_SOL_momentum_loss_fraction(1.0, 0.8, 2.1, target_electron_temp)

    elif key == MomentumLossFunction.Sang:
        return _calc_SOL_momentum_loss_fraction(1.3, 1.8, 1.6, target_electron_temp)

    elif key == MomentumLossFunction.Jarvinen:
        return _calc_SOL_momentum_loss_fraction(1.7, 2.2, 1.2, target_electron_temp)

    elif key == MomentumLossFunction.Moulton:
        return _calc_SOL_momentum_loss_fraction(1.0, 1.0, 1.5, target_electron_temp)

    elif key == MomentumLossFunction.PerezH:
        return _calc_SOL_momentum_loss_fraction(0.8, 2.0, 1.2, target_electron_temp)

    elif key == MomentumLossFunction.PerezL:
        return _calc_SOL_momentum_loss_fraction(1.1, 3.0, 0.9, target_electron_temp)

    else:
        raise NotImplementedError(f"No implementation for MomentumLossFunction {key}")



"""Reaction rates (<sigma*v>) as a function of ion temperature."""

import numpy as np
from numpy import float64
from numpy.typing import NDArray


def sigmav_DT(ion_temp_profile: NDArray[float64]) -> NDArray[float64]:
    r"""Deuterium-Tritium reaction.

    Calculate :math:`\langle \sigma v \rangle` for a given characteristic ion energy.
    Formulation from table 1, column S5 in :cite:`hively_convenient_1977`.
    Curvefit was performed for the range of [1,80]keV.

    Args:
        ion_temp_profile: ion temperature profile [keV]

    Returns:
        :math:`\langle \sigma v \rangle` in cm^3/s.
    """
    A = [-21.377692, -25.204054, -7.1013427 * 1e-2, 1.9375451 * 1e-4, 4.9246592 * 1e-6, -3.9836572 * 1e-8]
    r = 0.2935
    sigmav = np.exp(
        A[0] / ion_temp_profile**r
        + A[1]
        + A[2] * ion_temp_profile
        + A[3] * ion_temp_profile**2.0
        + A[4] * ion_temp_profile**3.0
        + A[5] * ion_temp_profile**4.0
    )
    return sigmav  # type: ignore[no-any-return] # [cm^3/s]


def sigmav_DT_BoschHale(ion_temp_profile: NDArray[float64]) -> NDArray[float64]:
    r"""Deuterium-Tritium reaction.

    Calculate :math:`\langle \sigma v \rangle` product for a given characteristic ion energy using Bosch Hale method.

    :func:`sigmav_DT_BoschHale` is more accurate than :func:`sigmav_DT` for ion_temp_profile > ~48.45 keV (estimate based on
    linear interp between errors found at available datapoints).
    Maximum error = 1.4% within range 50-1000 keV from available NRL data.

    Formulation from :cite:`bosch_improved_1992`

    Args:
        ion_temp_profile: ion temperature profile [keV]

    Returns:
        :math:`\langle \sigma v \rangle` in cm^3/s.

    """
    # Bosch Hale coefficients for DT reaction
    C = [0.0, 1.173e-9, 1.514e-2, 7.519e-2, 4.606e-3, 1.35e-2, -1.068e-4, 1.366e-5]
    B_G = 34.3827
    mr_c2 = 1124656

    theta = ion_temp_profile / (
        1
        - (ion_temp_profile * (C[2] + ion_temp_profile * (C[4] + ion_temp_profile * C[6])))
        / (1 + ion_temp_profile * (C[3] + ion_temp_profile * (C[5] + ion_temp_profile * C[7])))
    )
    eta = (B_G**2 / (4 * theta)) ** (1 / 3)
    sigmav = C[1] * theta * np.sqrt(eta / (mr_c2 * ion_temp_profile**3)) * np.exp(-3 * eta)
    return sigmav  # type: ignore[no-any-return] # [cm^3/s]


def sigmav_DD(ion_temp_profile: NDArray[float64]) -> tuple[NDArray[float64], NDArray[float64], NDArray[float64]]:
    r"""Deuterium-Deuterium reaction.

    Calculate :math:`\langle \sigma v \rangle` for a given characteristic ion energy.
    Formulation from column S5, in table 3 and 4 in :cite:`hively_convenient_1977`.
    Curvefit was performed for the range of [1,80]keV.

    Args:
        ion_temp_profile: ion temperature profile [keV]

    Returns:
        :math:`\langle \sigma v \rangle` tuple (total, D(d,p)T, D(d,n)3He) in cm^3/s.
    """
    a_1 = [
        -15.511891,
        -35.318711,
        -1.2904737 * 1e-2,
        2.6797766 * 1e-4,
        -2.9198685 * 1e-6,
        1.2748415 * 1e-8,
    ]  # For D(d,p)T
    r_1 = 0.3735
    a_2 = [
        -15.993842,
        -35.017640,
        -1.3689787 * 1e-2,
        2.7089621 * 1e-4,
        -2.9441547 * 1e-6,
        1.2841202 * 1e-8,
    ]  # For D(d,n)3He
    r_2 = 0.3725
    # Ti in units of keV, sigmav in units of cm^3/s
    sigmav_1: NDArray[float64] = np.exp(
        a_1[0] / ion_temp_profile**r_1
        + a_1[1]
        + a_1[2] * ion_temp_profile
        + a_1[3] * ion_temp_profile**2.0
        + a_1[4] * ion_temp_profile**3.0
        + a_1[5] * ion_temp_profile**4.0
    )
    sigmav_2: NDArray[float64] = np.exp(
        a_2[0] / ion_temp_profile**r_2
        + a_2[1]
        + a_2[2] * ion_temp_profile
        + a_2[3] * ion_temp_profile**2.0
        + a_2[4] * ion_temp_profile**3.0
        + a_2[5] * ion_temp_profile**4.0
    )
    sigmav_tot: NDArray[float64] = sigmav_1 + sigmav_2
    return sigmav_tot, sigmav_1, sigmav_2  # [cm^3/s]


def sigmav_DD_BoschHale(ion_temp_profile: NDArray[float64]) -> tuple[NDArray[float64], NDArray[float64], NDArray[float64]]:
    r"""Deuterium-Deuterium reaction.

    Calculate :math:`\langle \sigma v \rangle` product for a given characteristic ion energy using Bosch Hale method.

    Function tested on available data at [1, 2, 5, 10, 20, 50, 100] keV.
    Maximum error = 3.8% within range 5-50 keV and increases significantly outside of [5, 50] keV.

    Uses DD cross section formulation from :cite:`bosch_improved_1992`.

    Other form in :cite:`langenbrunner_analytic_2017`.

    Args:
        ion_temp_profile: ion temperature profile [keV]

    Returns:
        :math:`\langle \sigma v \rangle` tuple (total, D(d,p)T, D(d,n)3He) in cm^3/s.
    """
    # For D(d,n)3He
    cBH_1 = [((31.3970**2) / 4.0) ** (1.0 / 3.0), 5.65718e-12, 3.41e-03, 1.99e-03, 0, 1.05e-05, 0, 0]  # 3.72e-16,

    mc2_1 = 937814.0

    # For D(d,p)T
    cBH_2 = [((31.3970**2) / 4.0) ** (1.0 / 3.0), 5.43360e-12, 5.86e-03, 7.68e-03, 0, -2.96e-06, 0, 0]  # 3.57e-16,

    mc2_2 = 937814.0

    thetaBH_1 = ion_temp_profile / (
        1
        - (
            (cBH_1[2] * ion_temp_profile + cBH_1[4] * ion_temp_profile**2 + cBH_1[6] * ion_temp_profile**3)
            / (1 + cBH_1[3] * ion_temp_profile + cBH_1[5] * ion_temp_profile**2 + cBH_1[7] * ion_temp_profile**3)
        )
    )

    thetaBH_2 = ion_temp_profile / (
        1
        - (
            (cBH_2[2] * ion_temp_profile + cBH_2[4] * ion_temp_profile**2 + cBH_2[6] * ion_temp_profile**3)
            / (1 + cBH_2[3] * ion_temp_profile + cBH_2[5] * ion_temp_profile**2 + cBH_2[7] * ion_temp_profile**3)
        )
    )

    etaBH_1: float = cBH_1[0] / (thetaBH_1 ** (1.0 / 3.0))
    etaBH_2: float = cBH_2[0] / (thetaBH_2 ** (1.0 / 3.0))

    sigmav_1: NDArray[float64] = cBH_1[1] * thetaBH_1 * np.sqrt(etaBH_1 / (mc2_1 * (ion_temp_profile**3.0))) * np.exp(-3.0 * etaBH_1)
    sigmav_2: NDArray[float64] = cBH_2[1] * thetaBH_2 * np.sqrt(etaBH_2 / (mc2_2 * (ion_temp_profile**3.0))) * np.exp(-3.0 * etaBH_2)
    sigmav_tot: NDArray[float64] = sigmav_1 + sigmav_2

    return sigmav_tot, sigmav_1, sigmav_2  # [cm^3/s]


def sigmav_DHe3(ion_temp_profile: NDArray[float64]) -> NDArray[float64]:
    r"""Deuterium-Helium-3 reaction.

    Calculate :math:`\langle \sigma v \rangle` for a given characteristic ion energy.

    Function tested on available data at [1, 2, 5, 10, 20, 50, 100] keV.
    Maximum error = 8.4% within range 2-100 keV and should not be used outside range [2, 100] keV.

    Uses DD cross section formulation :cite:`bosch_improved_1992`.

    Args:
        ion_temp_profile: ion temperature profile [keV]

    Returns:
        :math:`\langle \sigma v \rangle` in cm^3/s.
    """
    # For He3(d,p)4He
    cBH_1 = [
        ((68.7508**2) / 4.0) ** (1.0 / 3.0),
        5.51036e-10,  # 3.72e-16,
        6.41918e-03,
        -2.02896e-03,
        -1.91080e-05,
        1.35776e-04,
        0,
        0,
    ]

    mc2_1 = 1124572.0

    thetaBH_1 = ion_temp_profile / (
        1
        - (
            (cBH_1[2] * ion_temp_profile + cBH_1[4] * ion_temp_profile**2 + cBH_1[6] * ion_temp_profile**3)
            / (1 + cBH_1[3] * ion_temp_profile + cBH_1[5] * ion_temp_profile**2 + cBH_1[7] * ion_temp_profile**3.0)
        )
    )

    etaBH_1: float = cBH_1[0] / (thetaBH_1 ** (1.0 / 3.0))

    sigmav: NDArray[float64] = cBH_1[1] * thetaBH_1 * np.sqrt(etaBH_1 / (mc2_1 * (ion_temp_profile**3.0))) * np.exp(-3.0 * etaBH_1)

    return sigmav  # [cm^3/s]


def sigmav_pB11(ion_temp_profile: NDArray[float64]) -> NDArray[float64]:
    r"""Proton (hydrogen)-Boron11 reaction.

    Calculate :math:`\langle \sigma v \rangle` for a given characteristic ion energy.

    Uses cross section from Nevins and Swain :cite:`nevins_thermonuclear_2000`.
    Updated cross sections in :cite:`sikora_new_2016`, and :cite:`putvinski_fusion_2019` are not in analytic form.

    Args:
        ion_temp_profile: characteristic ion energy in keV.

    Returns:
        :math:`\langle \sigma v \rangle` in cm^3/s.
    """
    # High temperature (T>60 keV)
    # For B11(p,alpha)alpha,alpha
    cBH_1 = [
        ((22589.0) / 4.0) ** (1.0 / 3.0),
        4.4467e-14,
        -5.9357e-02,
        2.0165e-01,
        1.0404e-03,
        2.7621e-03,
        -9.1653e-06,
        9.8305e-07,
    ]

    mc2_1 = 859526.0

    thetaBH_1 = ion_temp_profile / (
        1
        - (
            (cBH_1[2] * ion_temp_profile + cBH_1[4] * ion_temp_profile**2 + cBH_1[6] * ion_temp_profile**3)
            / (1 + cBH_1[3] * ion_temp_profile + cBH_1[5] * ion_temp_profile**2 + cBH_1[7] * ion_temp_profile**3)
        )
    )

    etaBH_1 = cBH_1[0] / (thetaBH_1 ** (1.0 / 3.0))

    sigmavNRhigh = (
        cBH_1[1] * thetaBH_1 * np.sqrt(etaBH_1 / (mc2_1 * (ion_temp_profile**3.0))) * np.exp(-3.0 * etaBH_1)
    ) * 1e6  # m3 to cm3

    # Low temperature (T<60 keV)
    E0 = ((17.81) ** (1.0 / 3.0)) * (ion_temp_profile ** (2.0 / 3.0))

    deltaE0 = 4.0 * np.sqrt(ion_temp_profile * E0 / 3.0)

    tau = (3.0 * E0) / ion_temp_profile

    Mp = 1.0  # *1.67e-27
    MB = 11.0  # *1.67e-27

    Mr = (Mp * MB) / (Mp + MB)

    C0 = 197.000 * 1e-25  # MeVb to kev/m^2
    C1 = 0.240 * 1e-25  # MeVb to kev/m^2
    C2 = 0.000231 * 1e-25  # MeVb to kev/m^2

    Seff = (
        C0 * (1 + (5.0 / (12.0 * tau)))
        + C1 * (E0 + (35.0 / 36.0) * ion_temp_profile)
        + C2 * (E0**2.0 + (89.0 / 36.0) * E0 * ion_temp_profile)
    )

    sigmavNRlow: NDArray[float64] = (
        np.sqrt(2 * ion_temp_profile / Mr) * ((deltaE0 * Seff) / (ion_temp_profile ** (2.0))) * np.exp(-tau)
    ) * 1e6  # m3 to cm3
    # 148 keV resonance
    sigmavR: NDArray[float64] = (
        (5.41e-21) * ((1.0 / ion_temp_profile) ** (3.0 / 2.0)) * np.exp(-148.0 / ion_temp_profile)
    ) * 1e6  # m3 to cm3
    sigmav: NDArray[float64] = sigmavNRhigh

    for i in range(len(ion_temp_profile)):
        if ion_temp_profile[i] < 60.0:  # keV
            sigmav[i] = sigmavNRlow[i] + sigmavR[i]
        elif (ion_temp_profile[i] > 60.0) and (ion_temp_profile[i] < 130):  # keV
            sigmav[i] = sigmavNRhigh[i] + sigmavR[i]

    return sigmav  # [cm^3/s]



"""Algorithm wrappers for single functions which don't fit into larger algorithms."""
import numpy as np

from .. import formulas
from ..named_options import Algorithms
from ..unit_handling import ureg
from .algorithm_class import Algorithm

calc_confinement_transition_threshold_power = Algorithm.from_single_function(
    formulas.calc_confinement_transition_threshold_power, return_keys=["P_LH_thresh"], name="calc_confinement_transition_threshold_power"
)
calc_ratio_P_LH = Algorithm.from_single_function(
    func=lambda P_sol, P_LH_thresh: P_sol / P_LH_thresh, return_keys=["ratio_of_P_SOL_to_P_LH"], name="calc_ratio_P_LH"
)
calc_f_rad_core = Algorithm.from_single_function(
    func=lambda P_radiation, P_in: P_radiation / P_in, return_keys=["core_radiated_power_fraction"], name="calc_f_rad_core"
)
calc_normalised_collisionality = Algorithm.from_single_function(
    func=formulas.calc_normalised_collisionality, return_keys=["nu_star"], name="calc_normalised_collisionality"
)
calc_rho_star = Algorithm.from_single_function(func=formulas.calc_rho_star, return_keys=["rho_star"], name="calc_rho_star")
calc_triple_product = Algorithm.from_single_function(
    func=formulas.calc_triple_product, return_keys=["fusion_triple_product"], name="calc_triple_product"
)
calc_greenwald_fraction = Algorithm.from_single_function(
    func=formulas.calc_greenwald_fraction, return_keys=["greenwald_fraction"], name="calc_greenwald_fraction"
)
calc_current_relaxation_time = Algorithm.from_single_function(
    func=formulas.calc_current_relaxation_time, return_keys=["current_relaxation_time"], name="calc_current_relaxation_time"
)
calc_peak_pressure = Algorithm.from_single_function(
    func=formulas.calc_peak_pressure, return_keys=["peak_pressure"], name="calc_peak_pressure"
)
calc_average_total_pressure = Algorithm.from_single_function(
    lambda average_electron_density, average_electron_temp, average_ion_temp: average_electron_density
    * (average_electron_temp + average_ion_temp),
    return_keys=["average_total_pressure"],
    name="calc_average_total_pressure",
)
calc_bootstrap_fraction = Algorithm.from_single_function(
    formulas.calc_bootstrap_fraction, return_keys=["bootstrap_fraction"], name="calc_bootstrap_fraction"
)
calc_auxillary_power = Algorithm.from_single_function(
    lambda P_external, P_ohmic: (P_external - P_ohmic).clip(min=0.0 * ureg.MW), return_keys=["P_auxillary"], name="calc_auxillary_power"
)
calc_average_ion_temp = Algorithm.from_single_function(
    lambda average_electron_temp, ion_to_electron_temp_ratio: average_electron_temp * ion_to_electron_temp_ratio,
    return_keys=["average_ion_temp"],
    name="calc_average_ion_temp",
)
calc_fuel_average_mass_number = Algorithm.from_single_function(
    formulas.calc_fuel_average_mass_number, return_keys=["fuel_average_mass_number"], name="calc_fuel_average_mass_number"
)
calc_magnetic_field_on_axis = Algorithm.from_single_function(
    lambda product_of_magnetic_field_and_radius, major_radius: product_of_magnetic_field_and_radius / major_radius,
    return_keys=["magnetic_field_on_axis"],
    name="calc_magnetic_field_on_axis",
)
require_P_rad_less_than_P_in = Algorithm.from_single_function(
    lambda P_in, P_radiation: np.minimum(P_radiation, P_in), return_keys=["P_radiation"], name="require_P_rad_less_than_P_in"
)
calc_P_SOL = Algorithm.from_single_function(
    lambda P_in, P_radiation: np.maximum(P_in - P_radiation, 0.0), return_keys=["P_sol"], name="calc_P_SOL"
)
calc_plasma_stored_energy = Algorithm.from_single_function(
    lambda average_electron_density, average_electron_temp, average_ion_density, summed_impurity_density, average_ion_temp, plasma_volume: (
        (3.0 / 2.0)
        * ((average_electron_density * average_electron_temp) + ((average_ion_density + summed_impurity_density) * average_ion_temp))
        * plasma_volume
    ).pint.to(ureg.MJ),
    return_keys=["plasma_stored_energy"],
    name="calc_plasma_stored_energy",
)

SINGLE_FUNCTIONS = {Algorithms[key]: val for key, val in locals().items() if isinstance(val, Algorithm)}



"""Common functionality shared between other functions."""
import numpy as np
from numpy import float64
from numpy.typing import NDArray


def integrate_profile_over_volume(
    array_per_m3: NDArray[float64],
    rho: NDArray[float64],
    plasma_volume: float,
) -> float:
    """Approximate the volume integral of a profile given as a function of rho.

    Args:
        array_per_m3: a profile of values [units * m^-3]
        rho: [~] :term:`glossary link<rho>`
        plasma_volume: [m^3] :term:`glossary link<plasma_volume>`

    Returns:
         volume_integrated_value [units]
    """
    drho = rho[1] - rho[0]
    return float(np.sum(array_per_m3 * 2.0 * rho * drho) * plasma_volume)



"""Calculate the power due to Ohmic resistive heating."""
from .. import formulas
from ..unit_handling import Unitfull, convert_to_default_units
from .algorithm_class import Algorithm

RETURN_KEYS = [
    "spitzer_resistivity",
    "trapped_particle_fraction",
    "neoclassical_loop_resistivity",
    "loop_voltage",
    "P_ohmic",
]


def run_calc_ohmic_power(
    bootstrap_fraction: Unitfull,
    average_electron_temp: Unitfull,
    inverse_aspect_ratio: Unitfull,
    z_effective: Unitfull,
    major_radius: Unitfull,
    minor_radius: Unitfull,
    areal_elongation: Unitfull,
    plasma_current: Unitfull,
) -> dict[str, Unitfull]:
    """Calculate the power due to Ohmic resistive heating.

    Args:
        bootstrap_fraction: :term:`glossary link<bootstrap_fraction>`
        average_electron_temp: :term:`glossary link<average_electron_temp>`
        inverse_aspect_ratio: :term:`glossary link<inverse_aspect_ratio>`
        z_effective: :term:`glossary link<z_effective>`
        major_radius: :term:`glossary link<major_radius>`
        minor_radius: :term:`glossary link<minor_radius>`
        areal_elongation: :term:`glossary link<areal_elongation>`
        plasma_current: :term:`glossary link<plasma_current>`

    Returns:
    :term:`spitzer_resistivity`, :term:`trapped_particle_fraction`, :term:`neoclassical_loop_resistivity`, :term:`loop_voltage`, :term:`P_ohmic`
    """
    inductive_plasma_current = plasma_current * (1.0 - bootstrap_fraction)
    spitzer_resistivity = formulas.calc_Spitzer_loop_resistivity(average_electron_temp)
    trapped_particle_fraction = formulas.calc_resistivity_trapped_enhancement(inverse_aspect_ratio)
    neoclassical_loop_resistivity = formulas.calc_neoclassical_loop_resistivity(spitzer_resistivity, z_effective, trapped_particle_fraction)
    loop_voltage = formulas.calc_loop_voltage(
        major_radius, minor_radius, inductive_plasma_current, areal_elongation, neoclassical_loop_resistivity
    )
    P_ohmic = formulas.calc_ohmic_power(inductive_plasma_current, loop_voltage)

    local_vars = locals()
    return {key: convert_to_default_units(local_vars[key], key) for key in RETURN_KEYS}


calc_ohmic_power = Algorithm(
    function=run_calc_ohmic_power,
    return_keys=RETURN_KEYS,
)



"""Routines to calculate the target electron flux, following the 2-point-model method of Stangeby, PPCF 2018."""
from typing import Union

import xarray as xr

from ...unit_handling import Quantity


def calc_target_electron_flux(
    target_electron_flux_basic: Union[Quantity, xr.DataArray],
    f_vol_loss_target_electron_flux: Union[float, xr.DataArray],
    f_other_target_electron_flux: Union[float, xr.DataArray],
) -> Union[Quantity, xr.DataArray]:
    """Calculate the target electron flux, correcting for volume-losses and other effects.

    Components are calculated using the other functions in this file.
    """
    return target_electron_flux_basic * f_vol_loss_target_electron_flux * f_other_target_electron_flux


def calc_target_electron_flux_basic(
    fuel_average_mass_number: Union[Quantity, xr.DataArray],
    parallel_heat_flux_density: Union[Quantity, xr.DataArray],
    upstream_total_pressure: Union[Quantity, xr.DataArray],
    sheath_heat_transmission_factor: Union[float, xr.DataArray],
) -> Union[Quantity, xr.DataArray]:
    """Calculate the flux of electrons (particles per square-metre per second) at the target according to the basic two-point-model.

    From equation 24, :cite:`stangeby_2018`.

    Args:
        fuel_average_mass_number: [amu]
        parallel_heat_flux_density: [GW/m^2]
        upstream_total_pressure: [atm]
        sheath_heat_transmission_factor: [~]

    Returns:
        target_electron_flux_basic [m^-2 s^-1]
    """
    return sheath_heat_transmission_factor / (8.0 * fuel_average_mass_number) * upstream_total_pressure**2 / parallel_heat_flux_density


def calc_f_vol_loss_target_electron_flux(
    SOL_power_loss_fraction: Union[float, xr.DataArray],
    SOL_momentum_loss_fraction: Union[float, xr.DataArray],
) -> Union[float, xr.DataArray]:
    """Calculate the volume-loss correction term for the electron flux at the target.

    From equation 24, :cite:`stangeby_2018`.

    Args:
        SOL_power_loss_fraction: f_cooling [~]
        SOL_momentum_loss_fraction: f_mom-loss [~]

    Returns:
        f_vol_loss_target_electron_flux [~]
    """
    return (1.0 - SOL_momentum_loss_fraction) ** 2 / (1.0 - SOL_power_loss_fraction)


def calc_f_other_target_electron_flux(
    target_ratio_of_ion_to_electron_temp: Union[float, xr.DataArray],
    target_ratio_of_electron_to_ion_density: Union[float, xr.DataArray],
    target_mach_number: Union[float, xr.DataArray],
    toroidal_flux_expansion: Union[float, xr.DataArray],
) -> Union[float, xr.DataArray]:
    """Calculate correction terms other than the volume-loss correction for the electron flux at the target.

    Includes flux expansion, dilution of ions, different electron and ion temperatures and sub/super-sonic outflow.
    From equation 24, :cite:`stangeby_2018`., with

    Args:
        target_ratio_of_ion_to_electron_temp: tau_t = (T_i / T_e)_target [equation 21] [~]
        target_ratio_of_electron_to_ion_density: z_t = (ne / total ion density)_target [equation 22] [~]
        target_mach_number: M_t = (parallel ion velocity / sound speed)_target [~]
        toroidal_flux_expansion: R_t/R_u = major radius at target / major radius upstream [see discussion around equation 12] [~]

    Returns:
        f_other_target_electron_flux [~]
    """
    return (
        (2.0 / (1.0 + target_ratio_of_ion_to_electron_temp / target_ratio_of_electron_to_ion_density))
        * 4.0
        * target_mach_number**2
        / (1.0 + target_mach_number**2) ** 2
        * toroidal_flux_expansion
    )



"""Calculate the thermal gain factor (Q, Q_plasma, Q_thermal)."""
import numpy as np

from ..unit_handling import ureg, wraps_ufunc

_IGNITED_THRESHOLD = 1e3
_IGNITED = 1e6


def _ignition_above_threshold(Q: float) -> float:
    """If Q > _IGNITED_THRESHOLD, set equal to _IGNITED.

    Args:
        Q: Fusion power gain [~]

    Returns:
         Q [~]
    """
    if Q > _IGNITED_THRESHOLD:
        return _IGNITED
    else:
        return Q


@wraps_ufunc(return_units=dict(Q=ureg.dimensionless), input_units=dict(P_fusion=ureg.MW, P_launched=ureg.MW))
def thermal_calc_gain_factor(P_fusion: float, P_launched: float) -> float:
    """Calculate the fusion gain.

    Args:
        P_fusion: [MW] :term:`glossary link<P_fusion>`
        P_launched: [MW] :term:`glossary link<P_launched>`

    Returns:
         Q [~]
    """
    if np.isclose(P_launched, 0.0):
        return _IGNITED
    else:
        return _ignition_above_threshold(P_fusion / P_launched)



"""Calculate the average fuel mass in atomic mass units."""
from typing import Callable

from ..named_options import ReactionType
from ..unit_handling import ureg, wraps_ufunc

FUEL_MASS_AMU: dict[ReactionType, Callable[[float], float]] = {
    ReactionType.DT: lambda heavier_fuel_species_fraction: 2.0 + heavier_fuel_species_fraction,
    ReactionType.DD: lambda _: 2.0,
    ReactionType.DHe3: lambda heavier_fuel_species_fraction: 2.0 + heavier_fuel_species_fraction,
    ReactionType.pB11: lambda heavier_fuel_species_fraction: 11.0 * heavier_fuel_species_fraction
    + 1.0 * (1.0 - heavier_fuel_species_fraction),
}


@wraps_ufunc(
    return_units=dict(fuel_average_mass_number=ureg.amu),
    input_units=dict(fusion_reaction=None, heavier_fuel_species_fraction=ureg.dimensionless),
)
def calc_fuel_average_mass_number(fusion_reaction: ReactionType, heavier_fuel_species_fraction: float) -> float:
    """Calculate the average mass of the fuel ions, based on reaction type and fuel mixture ratio.

    Args:
        fusion_reaction: reaction type.
        heavier_fuel_species_fraction: n_heavier / (n_heavier + n_lighter) number fraction.

    Returns:
        :term:`fuel_average_mass_number` [amu]
    """
    return FUEL_MASS_AMU[fusion_reaction](heavier_fuel_species_fraction)



"""Routines to calculate the upstream electron temperature."""

from typing import Union

import xarray as xr

from ...unit_handling import Quantity


def calc_upstream_electron_temp(
    target_electron_temp: Union[Quantity, xr.DataArray],
    parallel_heat_flux_density: Union[Quantity, xr.DataArray],
    parallel_connection_length: Union[Quantity, xr.DataArray],
    kappa_e0: Union[Quantity, xr.DataArray],
    SOL_conduction_fraction: Union[float, xr.DataArray] = 1.0,
) -> Union[Quantity, xr.DataArray]:
    """Calculate the upstream electron temperature.

    Equation 38 from :cite:`stangeby_2018`, keeping the dependence on target_electron_temp.

    Args:
        target_electron_temp: [eV]
        parallel_heat_flux_density: [GW/m^2]
        parallel_connection_length: [m]
        kappa_e0: [W / (eV**3.5 m)]
        SOL_conduction_fraction: [eV]

    Returns:
        upstream_electron_temp [eV]
    """
    return (
        target_electron_temp**3.5 + 3.5 * (SOL_conduction_fraction * parallel_heat_flux_density * parallel_connection_length / kappa_e0)
    ) ** (2.0 / 7.0)



"""Calculate the parallel heat flux density upstream and related metrics."""
import numpy as np

from .. import formulas, named_options
from ..unit_handling import Unitfull, convert_to_default_units, ureg
from .algorithm_class import Algorithm

RETURN_KEYS = [
    "PB_over_R",
    "PBpRnSq",
    "B_pol_out_mid",
    "B_t_out_mid",
    "fieldline_pitch_at_omp",
    "lambda_q",
    "q_parallel",
    "q_perp",
]


def run_calc_heat_exhaust(
    P_sol: Unitfull,
    magnetic_field_on_axis: Unitfull,
    major_radius: Unitfull,
    inverse_aspect_ratio: Unitfull,
    plasma_current: Unitfull,
    minor_radius: Unitfull,
    q_star: Unitfull,
    average_electron_density: Unitfull,
    average_total_pressure: Unitfull,
    fraction_of_P_SOL_to_divertor: Unitfull,
    lambda_q_scaling: named_options.LambdaQScaling,
    lambda_q_factor: Unitfull = 1.0 * ureg.dimensionless,
) -> dict[str, Unitfull]:
    """Calculate the parallel heat flux density upstream and related metrics.

    Args:
        P_sol: :term:`glossary link<P_sol>`
        magnetic_field_on_axis: :term:`glossary link<magnetic_field_on_axis>`
        major_radius: :term:`glossary link<major_radius>`
        inverse_aspect_ratio: :term:`glossary link<inverse_aspect_ratio>`
        plasma_current: :term:`glossary link<plasma_current>`
        minor_radius: :term:`glossary link<minor_radius>`
        q_star: :term:`glossary link<q_star>`
        average_electron_density: :term:`glossary link<average_electron_density>`
        average_total_pressure: :term:`glossary link <average_total_pressure>`
        fraction_of_P_SOL_to_divertor: :term:`glossary link <fraction_of_P_SOL_to_divertor>`
        lambda_q_scaling: :term:`glossary link<lambda_q_scaling>`
        lambda_q_factor: :term:`glossary link<lambda_q_factor>`

    Returns:
        :term:`PB_over_R`, :term:`PBpRnSq`, :term:`B_pol_out_mid`, :term:`B_t_out_mid`, :term:`fieldline_pitch_at_omp`, :term:`lambda_q`, :term:`q_parallel`, :term:`q_perp`

    """
    PB_over_R = P_sol * magnetic_field_on_axis / major_radius
    PBpRnSq = (P_sol * (magnetic_field_on_axis / q_star) / major_radius) / (average_electron_density**2.0)

    B_pol_out_mid = formulas.calc_B_pol_omp(plasma_current=plasma_current, minor_radius=minor_radius)
    B_t_out_mid = formulas.calc_B_tor_omp(magnetic_field_on_axis, major_radius, minor_radius)

    fieldline_pitch_at_omp = np.sqrt(B_t_out_mid**2 + B_pol_out_mid**2) / B_pol_out_mid

    lambda_q = lambda_q_factor * formulas.scrape_off_layer_model.calc_lambda_q(
        lambda_q_scaling, average_total_pressure, P_sol, major_radius, B_pol_out_mid, inverse_aspect_ratio
    )

    q_parallel = formulas.scrape_off_layer_model.calc_parallel_heat_flux_density(
        P_sol, fraction_of_P_SOL_to_divertor, major_radius + minor_radius, lambda_q, fieldline_pitch_at_omp
    )
    q_perp = P_sol / (2.0 * np.pi * (major_radius + minor_radius) * lambda_q)

    local_vars = locals()
    return {key: convert_to_default_units(local_vars[key], key) for key in RETURN_KEYS}


calc_heat_exhaust = Algorithm(
    function=run_calc_heat_exhaust,
    return_keys=RETURN_KEYS,
)



"""Enumerators to constrain options for functions."""
from enum import Enum, auto


class Algorithms(Enum):
    """Select which top-level algorithm to run."""

    predictive_popcon = auto()
    two_point_model_fixed_fpow = auto()
    two_point_model_fixed_qpart = auto()
    two_point_model_fixed_tet = auto()
    calc_beta = auto()
    calc_core_radiated_power = auto()
    calc_fusion_gain = auto()
    calc_geometry = auto()
    calc_heat_exhaust = auto()
    calc_ohmic_power = auto()
    calc_peaked_profiles = auto()
    calc_plasma_current_from_q_star = auto()
    calc_q_star_from_plasma_current = auto()
    calc_power_balance_from_tau_e = auto()
    calc_zeff_and_dilution_from_impurities = auto()
    calc_confinement_transition_threshold_power = auto()
    calc_ratio_P_LH = auto()
    calc_f_rad_core = auto()
    calc_normalised_collisionality = auto()
    calc_rho_star = auto()
    calc_triple_product = auto()
    calc_greenwald_fraction = auto()
    calc_current_relaxation_time = auto()
    calc_peak_pressure = auto()
    calc_average_total_pressure = auto()
    calc_bootstrap_fraction = auto()
    calc_auxillary_power = auto()
    calc_average_ion_temp = auto()
    calc_fuel_average_mass_number = auto()
    calc_magnetic_field_on_axis = auto()
    calc_extrinsic_core_radiator = auto()
    require_P_rad_less_than_P_in = auto()
    calc_P_SOL = auto()
    use_LOC_tau_e_below_threshold = auto()
    calc_plasma_stored_energy = auto()


class ProfileForm(Enum):
    """Methods to calculate nT profiles."""

    analytic = auto()
    prf = auto()


class RadiationMethod(Enum):
    """Methods to calculate radiation losses."""

    Inherent = "Bremsstrahlung and synchrotron radiation only"
    PostJensen = "Impurity radiation, using a coronal equilibrium model from Post & Jensen 1977"
    MavrinCoronal = "Impurity radiation, using a coronal equilibrium model from Mavrin 2018"
    MavrinNoncoronal = "Impurity radiation, using a non-coronal model from Mavrin 2017"
    Radas = "Impurity line and bremsstrahlung radiation, using coronal Lz curves from Radas"


class ReactionType(Enum):
    """Supported Fusion Fuel Reaction Types."""

    DT = "Deuterium-Tritium"
    DD = "Deuterium-Deuterium"
    DHe3 = "Deuterium-Helium3"
    pB11 = "Proton-Boron11"


class Impurity(Enum):
    """Enum of possible impurity elements.

    The enum value represents the element's atomic number (Z).
    """

    Helium = 2
    Lithium = 3
    Beryllium = 4
    Carbon = 6
    Nitrogen = 7
    Oxygen = 8
    Neon = 10
    Argon = 18
    Krypton = 36
    Xenon = 54
    Tungsten = 74


class ConfinementScaling(Enum):
    r"""Enum of implemented \tau_{E} scalings."""
    ITER98y2 = auto()
    ITER89P = auto()
    ITER89P_ka = auto()
    ITERL96Pth = auto()
    ITER97L = auto()
    IModey2 = auto()
    ITPA20_STD5 = auto()
    ITPA20_IL = auto()
    ITPA20_IL_HighZ = auto()
    ITPA_2018_STD5_OLS = auto()
    ITPA_2018_STD5_WLS = auto()
    ITPA_2018_STD5_GLS = auto()
    ITPA_2018_STD5_SEL1_OLS = auto()
    ITPA_2018_STD5_SEL1_WLS = auto()
    ITPA_2018_STD5_SEL1_GLS = auto()
    LOC = auto()
    H_DS03 = auto()


class MomentumLossFunction(Enum):
    """Select which SOL momentum loss function to use."""

    KotovReiter = auto()
    Sang = auto()
    Jarvinen = auto()
    Moulton = auto()
    PerezH = auto()
    PerezL = auto()


class LambdaQScaling(Enum):
    """Options for heat flux decay length scaling."""

    Brunner = auto()
    EichRegression14 = auto()
    EichRegression15 = auto()



"""Calculate the radiated power due to impurities, according to an analytical fitted curve from Mavrin 2018."""

import warnings

import numpy as np
from numpy import float64
from numpy.polynomial.polynomial import polyval
from numpy.typing import NDArray

from ...named_options import Impurity
from ...unit_handling import ureg, wraps_ufunc
from ..helpers import integrate_profile_over_volume


@wraps_ufunc(
    return_units=dict(radiated_power=ureg.MW),
    input_units=dict(
        rho=ureg.dimensionless,
        electron_temp_profile=ureg.keV,
        electron_density_profile=ureg.n19,
        impurity_concentration=ureg.dimensionless,
        impurity_species=None,
        plasma_volume=ureg.m**3,
    ),
    input_core_dims=[("dim_rho",), ("dim_rho",), ("dim_rho",), (), (), ()],
)
def calc_impurity_radiated_power_mavrin_coronal(  # noqa: PLR0912, PLR0915
    rho: NDArray[float64],
    electron_temp_profile: NDArray[float64],
    electron_density_profile: NDArray[float64],
    impurity_concentration: float,
    impurity_species: Impurity,
    plasma_volume: float,
) -> float:
    """Calculation of radiated power, using fits from A.A. Mavrin's 2018 paper.

    "Improved fits of coronal radiative cooling rates for high-temperature plasmas."

    :cite:`mavrin_improved_2018`

    Args:
        rho: [~] :term:`glossary link<rho>`
        electron_temp_profile: [keV] :term:`glossary link<electron_temp_profile>`
        electron_density_profile: [1e19 m^-3] :term:`glossary link<electron_density_profile>`
        impurity_species: [] :term:`glossary link<impurity_species>`
        impurity_concentration: [~] :term:`glossary link<impurity_concentration>`
        plasma_volume: [m^3] :term:`glossary link<plasma_volume>`

    Returns:
      k  [MW] Estimated radiation power due to this impurity
    """
    impurity_Z = impurity_species.value

    zimp = np.array([2, 3, 4, 6, 7, 8, 10, 18, 36, 54, 74])

    if impurity_Z not in zimp:  # pragma: no cover
        warnings.warn(f"Mavrin 2018 line radiation calculation not supported for impurity with Z={impurity_Z}", stacklevel=3)
        return np.nan

    # If trying to evaluate for a temperature outside of the given range, assume nearest neighbor
    # and throw a warning
    if any(electron_density_profile < 0.1) or any(electron_density_profile > 100):  # pragma: no cover
        warnings.warn(
            "Mavrin 2018 line radiation calculation is only valid between 0.1-100keV. Using nearest neighbor extrapolation.", stacklevel=3
        )
    electron_density_profile = np.maximum(electron_density_profile, 0.1)
    electron_density_profile = np.minimum(electron_density_profile, 100)

    # L_z coefficients for the 11 supported impurities
    if impurity_Z == 2:  # Helium
        temperature_bin_borders = np.array([0.0, 100.0])
        radc = np.array(
            [
                [-3.5551e01, 3.1469e-01, 1.0156e-01, -9.3730e-02, 2.5020e-02],
            ]
        )

    elif impurity_Z == 3:  # Lithium
        temperature_bin_borders = np.array([0.0, 100.0])
        radc = np.array(
            [
                [-3.5115e01, 1.9475e-01, 2.5082e-01, -1.6070e-01, 3.5190e-02],
            ]
        )

    elif impurity_Z == 4:  # Beryllium
        temperature_bin_borders = np.array([0.0, 100.0])
        radc = np.array(
            [
                [-3.4765e01, 3.7270e-02, 3.8363e-01, -2.1384e-01, 4.1690e-02],
            ]
        )

    elif impurity_Z == 6:  # Carbon
        temperature_bin_borders = np.array([0.0, 0.5, 100.0])
        radc = np.array(
            [
                [-3.4738e01, -5.0085e00, -1.2788e01, -1.6637e01, -7.2904e00],
                [-3.4174e01, -3.6687e-01, 6.8856e-01, -2.9191e-01, 4.4470e-02],
            ]
        )

    elif impurity_Z == 7:  # Nitrogen
        temperature_bin_borders = np.array([0.0, 0.5, 2.0, 100.0])
        radc = np.array(
            [
                [-3.4065e01, -2.3614e00, -6.0605e00, -1.1570e01, -6.9621e00],
                [-3.3899e01, -5.9668e-01, 7.6272e-01, -1.7160e-01, 5.8770e-02],
                [-3.3913e01, -5.2628e-01, 7.0047e-01, -2.2790e-01, 2.8350e-02],
            ]
        )

    elif impurity_Z == 8:  # Oxygen
        temperature_bin_borders = np.array([0.0, 0.3, 100.0])
        radc = np.array(
            [
                [-3.7257e01, -1.5635e01, -1.7141e01, -5.3765e00, 0.0000e00],
                [-3.3640e01, -7.6211e-01, 7.9655e-01, -2.0850e-01, 1.4360e-02],
            ]
        )

    elif impurity_Z == 10:  # Neon
        temperature_bin_borders = np.array([0.0, 0.7, 5, 100.0])
        radc = np.array(
            [
                [-3.3132e01, 1.7309e00, 1.5230e01, 2.8939e01, 1.5648e01],
                [-3.3290e01, -8.7750e-01, 8.6842e-01, -3.9544e-01, 1.7244e-01],
                [-3.3410e01, -4.5345e-01, 2.9731e-01, 4.3960e-02, -2.6930e-02],
            ]
        )

    elif impurity_Z == 18:  # Argon
        temperature_bin_borders = np.array([0.0, 0.6, 3, 100.0])
        radc = np.array(
            [
                [-3.2155e01, 6.5221e00, 3.0769e01, 3.9161e01, 1.5353e01],
                [-3.2530e01, 5.4490e-01, 1.5389e00, -7.6887e00, 4.9806e00],
                [-3.1853e01, -1.6674e00, 6.1339e-01, 1.7480e-01, -8.2260e-02],
            ]
        )

    elif impurity_Z == 36:  # Krypton
        temperature_bin_borders = np.array([0.0, 0.447, 2.364, 100.0])
        radc = np.array(
            [
                [-3.4512e01, -2.1484e01, -4.4723e01, -4.0133e01, -1.3564e01],
                [-3.1399e01, -5.0091e-01, 1.9148e00, -2.5865e00, -5.2704e00],
                [-2.9954e01, -6.3683e00, 6.6831e00, -2.9674e00, 4.8356e-01],
            ]
        )

    elif impurity_Z == 54:  # Xenon
        temperature_bin_borders = np.array([0.0, 0.5, 2.5, 10, 100.0])
        radc = np.array(
            [
                [-2.9303e01, 1.4351e01, 4.7081e01, 5.9580e01, 2.5615e01],
                [-3.1113e01, 5.9339e-01, 1.2808e00, -1.1628e01, 1.0748e01],
                [-2.5813e01, -2.7526e01, 4.8614e01, -3.6885e01, 1.0069e01],
                [-2.2138e01, -2.2592e01, 1.9619e01, -7.5181e00, 1.0858e00],
            ]
        )

    elif impurity_Z == 74:  # Tungsten
        temperature_bin_borders = np.array([0.0, 1.5, 4, 100.0])
        radc = np.array(
            [
                [-3.0374e01, 3.8304e-01, -9.5126e-01, -1.0311e00, -1.0103e-01],
                [-3.0238e01, -2.9208e00, 2.2824e01, -6.3303e01, 5.1849e01],
                [-3.2153e01, 5.2499e00, -6.2740e00, 2.6627e00, -3.6759e-01],
            ]
        )

    # solve for radiated power

    Tlog = np.log10(electron_density_profile)
    log10_Lz = np.zeros(electron_density_profile.size)

    for i in range(len(radc)):
        it = np.nonzero(
            (electron_density_profile >= temperature_bin_borders[i]) & (electron_density_profile < temperature_bin_borders[i + 1])
        )[0]
        if it.size > 0:
            log10_Lz[it] = polyval(Tlog[it], radc[i])  # type: ignore[no-untyped-call]

    radrate = 10.0**log10_Lz
    radrate[np.isnan(radrate)] = 0

    # 1e38 factor to account for the fact that our n_e values are electron_density_profile values
    qRad = radrate * electron_temp_profile * electron_temp_profile * impurity_concentration * 1e38  # W / (m^3 s)
    radiated_power = integrate_profile_over_volume(qRad, rho, plasma_volume)  # [W]

    return float(radiated_power) / 1e6  # MW



"""Constructors and helper functions."""
from typing import Any, Union

import xarray as xr

from .named_options import (
    Algorithms,
    ConfinementScaling,
    Impurity,
    LambdaQScaling,
    MomentumLossFunction,
    ProfileForm,
    RadiationMethod,
    ReactionType,
)


def convert_named_options(key: str, val: Any) -> Any:  # noqa: PLR0911, PLR0912
    """Given a 'key' matching a named_option, return the corresponding Enum value."""
    if key == "algorithms":
        return Algorithms[val]
    elif key == "energy_confinement_scaling":
        return ConfinementScaling[val]
    elif key == "profile_form":
        return ProfileForm[val]
    elif key == "radiated_power_method":
        return RadiationMethod[val]
    elif key == "fusion_reaction":
        return ReactionType[val]
    elif key == "impurity":
        return Impurity[val]
    elif key == "impurities":
        return make_impurities_array(list(val.keys()), list(val.values()))
    elif key == "core_radiator":
        return Impurity[val]
    elif key == "lambda_q_scaling":
        return LambdaQScaling[val]
    elif key == "SOL_momentum_loss_function":
        return MomentumLossFunction[val]
    elif key == "tauE_scaling":
        return ConfinementScaling[val]
    elif key == "reaction_type":
        return ReactionType[val]
    elif key == "radiation_method":
        return RadiationMethod[val]
    else:
        # If the key doesn't match, don't convert the value
        return val


def make_impurities_array(
    species_list: Union[list[Union[str, Impurity]], Union[str, Impurity]],
    concentrations_list: Union[list[Union[float, xr.DataArray]], Union[float, xr.DataArray]],
) -> xr.DataArray:
    """Make an xr.DataArray with impurity species and their corresponding concentrations.

    This array should be used as the `impurities` variable.
    """
    # Convert DataArrays of species into plain lists. This is useful if you want to store Impurity objects in a dataset.
    if isinstance(species_list, (xr.DataArray)):
        species_list = species_list.values.tolist()
    # Deal with single-value input (not recommended, but avoids a confusing user error)
    if isinstance(species_list, (str, Impurity)):
        species_list = [
            species_list,
        ]
    if isinstance(concentrations_list, (float, xr.DataArray)):
        concentrations_list = [
            concentrations_list,
        ]

    if not len(species_list) == len(concentrations_list):
        raise ValueError(f"Dimension mismatch. Input was species list [{species_list}], concentrations list [{concentrations_list}]")

    array = xr.DataArray()
    for species, concentration in zip(species_list, concentrations_list):
        array = extend_impurities_array(array, species, concentration)

    return array


def make_impurities_array_from_kwargs(**kwargs: Any) -> xr.DataArray:
    """Make an xr.DataArray with impurity species and their corresponding concentrations, using the format (species1=concentration1, ...)."""
    return make_impurities_array(list(kwargs.keys()), list(kwargs.values()))


def extend_impurities_array(array: xr.DataArray, species: Union[str, Impurity], concentration: Union[float, xr.DataArray]) -> xr.DataArray:
    """Append a new element to the impurities array.

    This method automatically handles broadcasting.

    N.b. You can also 'extend' an empty array, constructed via xr.DataArray()
    """
    if not isinstance(species, Impurity):
        species = Impurity[species.capitalize()]

    if not isinstance(concentration, xr.DataArray):
        concentration = xr.DataArray(concentration)
    concentration = concentration.expand_dims("dim_species").assign_coords(dim_species=[species])

    if array.ndim == 0:
        return concentration
    else:
        other_species = array.sel(dim_species=[s for s in array.dim_species if s != species])
        return xr.concat((other_species, concentration), dim="dim_species")



"""Empirical scalings for energy confinement time."""

from .tau_e_from_Wp import calc_tau_e_and_P_in_from_scaling, get_datasets, load_metadata

__all__ = [
    "calc_tau_e_and_P_in_from_scaling",
    "load_metadata",
    "get_datasets",
]



"""Reactions rates and power densities for various fusion reactions."""
from typing import Callable, Union

from numpy import float64
from numpy.typing import NDArray

from ...named_options import ReactionType
from .reaction_energies import reaction_energy_DD, reaction_energy_DHe3, reaction_energy_DT, reaction_energy_pB11
from .reaction_rate_coefficients import sigmav_DD, sigmav_DD_BoschHale, sigmav_DHe3, sigmav_DT, sigmav_DT_BoschHale, sigmav_pB11

SIGMAV: dict[
    ReactionType,
    Union[
        Callable[[NDArray[float64]], NDArray[float64]],
        Callable[[NDArray[float64]], tuple[NDArray[float64], NDArray[float64], NDArray[float64]]],
    ],
] = {
    ReactionType.DT: sigmav_DT,
    ReactionType.DD: sigmav_DD_BoschHale,
    ReactionType.DHe3: sigmav_DHe3,
    ReactionType.pB11: sigmav_pB11,
}

ENERGY: dict[
    ReactionType,
    Union[
        Callable[[NDArray[float64], float], tuple[float, float, float, NDArray[float64], NDArray[float64], NDArray[float64]]],
        Callable[
            [tuple[NDArray[float64], NDArray[float64], NDArray[float64]]],
            tuple[NDArray[float64], float, NDArray[float64], NDArray[float64], NDArray[float64], NDArray[float64]],
        ],
    ],
] = {
    ReactionType.DT: reaction_energy_DT,
    ReactionType.DD: reaction_energy_DD,
    ReactionType.DHe3: reaction_energy_DHe3,
    ReactionType.pB11: reaction_energy_pB11,
}

__all__ = [
    "SIGMAV",
    "ENERGY",
    "reaction_energy_DD",
    "reaction_energy_DHe3",
    "reaction_energy_DT",
    "reaction_energy_pB11",
    "sigmav_DD_BoschHale",
    "sigmav_DHe3",
    "sigmav_DT",
    "sigmav_pB11",
    "sigmav_DD",
    "sigmav_DT_BoschHale",
]



"""Calculate fusion power and corresponding neutron wall loading."""

from typing import Union

from numpy import float64
from numpy.typing import NDArray

from ..named_options import ReactionType
from ..unit_handling import ureg, wraps_ufunc
from .fusion_reaction_data import ENERGY, SIGMAV
from .helpers import integrate_profile_over_volume


@wraps_ufunc(
    return_units=dict(P_fusion=ureg.MW, P_neutron=ureg.MW, P_alpha=ureg.MW),
    input_units=dict(
        fusion_reaction=None,
        ion_temp_profile=ureg.keV,
        heavier_fuel_species_fraction=ureg.dimensionless,
        nfuel19=ureg.n19,
        rho=ureg.dimensionless,
        plasma_volume=ureg.m**3,
    ),
    input_core_dims=[(), ("dim_rho",), (), ("dim_rho",), ("dim_rho",), ()],
    output_core_dims=[(), (), ()],
)
def calc_fusion_power(
    fusion_reaction: ReactionType,
    ion_temp_profile: NDArray[float64],
    heavier_fuel_species_fraction: float,
    nfuel19: NDArray[float64],
    rho: NDArray[float64],
    plasma_volume: float,
) -> tuple[float, float, float]:
    """Calculate the fusion power.

    Args:
        fusion_reaction: which nuclear reaction is being considered
        ion_temp_profile: [keV] :term:`glossary link<ion_temp_profile>`
        heavier_fuel_species_fraction: [~] fraction of fuel mixture which is the heavier nuclide
        nfuel19: [1e19 m^-3] average fuel density
        rho: [~] :term:`glossary link<rho>`
        plasma_volume: [m^3] :term:`glossary link<plasma_volume>`

    Returns:
        :term:`P_fusion` [MW], :term:`P_neutron` [MW], :term:`P_alpha` [MW]
    """
    reaction_at_Ti = _calc_fusion_reaction_rate(fusion_reaction, ion_temp_profile, heavier_fuel_species_fraction)

    power_density_factor_MW_m3 = reaction_at_Ti[4]
    neutral_power_density_factor_MW_m3 = reaction_at_Ti[5]
    charged_power_density_factor_MW_m3 = reaction_at_Ti[6]

    total_fusion_power_MW = _integrate_power(
        power_density_factor_MW_m3=power_density_factor_MW_m3,
        fuel_density_per_m3=nfuel19 * 1e19,
        rho=rho,
        plasma_volume=plasma_volume,
    )

    fusion_power_to_neutral_MW = _integrate_power(
        power_density_factor_MW_m3=neutral_power_density_factor_MW_m3,
        fuel_density_per_m3=nfuel19 * 1e19,
        rho=rho,
        plasma_volume=plasma_volume,
    )

    fusion_power_to_charged_MW = _integrate_power(
        power_density_factor_MW_m3=charged_power_density_factor_MW_m3,
        fuel_density_per_m3=nfuel19 * 1e19,
        rho=rho,
        plasma_volume=plasma_volume,
    )

    return total_fusion_power_MW, fusion_power_to_neutral_MW, fusion_power_to_charged_MW


@wraps_ufunc(
    return_units=dict(neutron_power_flux_to_walls=ureg.MW / ureg.m**2, neutron_rate=ureg.s**-1),
    input_units=dict(
        P_neutron=ureg.MW,
        surface_area=ureg.m**2,
        fusion_reaction=None,
        ion_temp_profile=ureg.keV,
        heavier_fuel_species_fraction=ureg.dimensionless,
    ),
    input_core_dims=[(), (), (), ("dim_rho",), ()],
    output_core_dims=[(), ()],
)
def calc_neutron_flux_to_walls(
    P_neutron: float,
    surface_area: float,
    fusion_reaction: ReactionType,
    ion_temp_profile: NDArray[float64],
    heavier_fuel_species_fraction: float,
) -> tuple[float, float]:
    """Calculate the neutron loading on the wall.

    Args:
        P_neutron: [MW] :term:`glossary link<P_neutron>`
        surface_area: [m^2] :term:`glossary link<surface_area>`
        fusion_reaction: which nuclear reaction is being considered
        ion_temp_profile: [keV] :term:`glossary link<ion_temp_profile>`
        heavier_fuel_species_fraction: fraction of fuel mixture which is the heavier nuclide

    Returns:
        neutron_power_flux_to_walls [MW / m^2], neutron_rate [s^-1]
    """
    neutron_power_flux_to_walls = P_neutron / surface_area
    rxn_energy_neut = _calc_fusion_reaction_rate(fusion_reaction, ion_temp_profile, heavier_fuel_species_fraction)[2]
    if rxn_energy_neut > 0:  # This will happen for D-He3 reactions
        neutron_rate = P_neutron / rxn_energy_neut  # [MW / MJ] -> [1 / s]
    else:
        neutron_rate = 0.0

    return neutron_power_flux_to_walls, neutron_rate


@wraps_ufunc(
    return_units=dict(
        sigmav=ureg.cm**3 / ureg.s,
        rxn_energy=ureg.MJ,
        rxn_energy_neut=ureg.MJ,
        rxn_energy_charged=ureg.MJ,
        number_power_dens=ureg.MW * ureg.m**3,
        number_power_dens_neut=ureg.MW * ureg.m**3,
        number_power_dens_charged=ureg.MW * ureg.m**3,
    ),
    input_units=dict(
        fusion_reaction=None,
        ion_temp_profile=ureg.keV,
        heavier_fuel_species_fraction=ureg.dimensionless,
    ),
    input_core_dims=[(), ("dim_rho",), ()],
    output_core_dims=[("dim_rho",), (), (), (), ("dim_rho",), ("dim_rho",), ("dim_rho",)],
)
def calc_fusion_reaction_rate(
    fusion_reaction: ReactionType, ion_temp_profile: NDArray[float64], heavier_fuel_species_fraction: float
) -> tuple[
    NDArray[float64],
    Union[NDArray[float64], float],
    float,
    Union[NDArray[float64], float],
    NDArray[float64],
    NDArray[float64],
    NDArray[float64],
]:
    """Calculate reaction properties based on reaction type, mixture ratio, and temperature.

    Args:
        fusion_reaction: which nuclear reaction is being considered
        ion_temp_profile: [keV] :term:`glossary link<ion_temp_profile>`
        heavier_fuel_species_fraction: fraction of fuel mixture which is the heavier nuclide

    Returns:
        :A tuple holding:

        :sigmav: rate coefficient <sigma*v> for the given ion temperature [cm^3/s]
        :rxn_energy: total energy released per reaction [MJ]
        :rxn_energy_neut: energy released to neutral products per reaction [MJ]
        :rxn_energy_charged: energy released to charged products per reaction [MJ]
        :number_power_dens: power per unit volume divided by reactant densities [MW*m^3]
        :number_power_dens_neut: power per unit volume divided by reactant densities deposited in neutral products [MW*m^3]
        :number_power_dens_charged: power per unit volume divided by reactant densities deposited in charged products [MW*m^3]
    """
    return _calc_fusion_reaction_rate(fusion_reaction, ion_temp_profile, heavier_fuel_species_fraction)


def _calc_fusion_reaction_rate(
    fusion_reaction: ReactionType, ion_temp_profile: NDArray[float64], heavier_fuel_species_fraction: float
) -> tuple[
    NDArray[float64],
    Union[NDArray[float64], float],
    float,
    Union[NDArray[float64], float],
    NDArray[float64],
    NDArray[float64],
    NDArray[float64],
]:
    """Calculate reaction properties based on reaction type, mixture ratio, and temperature, without unit-handling.

    Args:
        fusion_reaction: which nuclear reaction is being considered
        ion_temp_profile: [keV] :term:`glossary link<ion_temp_profile>`
        heavier_fuel_species_fraction: fraction of fuel mixture which is the heavier nuclide

    Returns:
        :A tuple holding:

        :sigmav: rate coefficient <sigma*v> for the given ion temperature [cm^3/s]
        :rxn_energy: total energy released per reaction [MJ]
        :rxn_energy_neut: energy released to neutral products per reaction [MJ]
        :rxn_energy_charged: energy released to charged products per reaction [MJ]
        :number_power_dens: power per unit volume divided by reactant densities [MW*m^3]
        :number_power_dens_neut: power per unit volume divided by reactant densities deposited in neutral products [MW*m^3]
        :number_power_dens_charged: power per unit volume divided by reactant densities deposited in charged products [MW*m^3]
    """
    sigmav_func = SIGMAV[fusion_reaction]  # Reaction rate function to use based on reaction type
    energy_func = ENERGY[fusion_reaction]  # Reaction energy function to use based on reaction type

    sigmav = sigmav_func(ion_temp_profile)

    # This generates a false positive when type checking, as the type checker doesn't
    # realize that the sigmav_func and energy_func pair always correctly matches.
    # That's because the return type of a dictionary can't be narrowed based on a runtime key.
    (rxn_energy, rxn_energy_neut, rxn_energy_charged, number_power_dens, number_power_dens_neut, number_power_dens_charged,) = energy_func(
        sigmav=sigmav, heavier_fuel_species_fraction=heavier_fuel_species_fraction  # type:ignore[call-arg]
    )

    return (
        sigmav,
        rxn_energy,
        rxn_energy_neut,
        rxn_energy_charged,
        number_power_dens,
        number_power_dens_neut,
        number_power_dens_charged,
    )  # type:ignore[return-value]


def _integrate_power(
    power_density_factor_MW_m3: NDArray[float64],
    fuel_density_per_m3: NDArray[float64],
    rho: NDArray[float64],
    plasma_volume: float,
) -> float:
    """Calculate the total power due to a nuclear reaction.

    Args:
        power_density_factor_MW_m3: energy per unit volume divided by fuel species densities [MW*m^3]
        fuel_density_per_m3: density of fuel species [m^-3]
        rho: [~] :term:`glossary link<rho>`
        plasma_volume: [m^3] :term:`glossary link<plasma_volume>`

    Returns:
         power [MW]
    """
    power_density_MW_per_m3 = power_density_factor_MW_m3 * fuel_density_per_m3 * fuel_density_per_m3
    power_MW = integrate_profile_over_volume(power_density_MW_per_m3, rho, plasma_volume)

    return power_MW



"""Calculate the effect of impurities on the effective charge and dilution."""
import numpy as np
import xarray as xr

from ..named_options import Impurity
from ..unit_handling import ureg, wraps_ufunc


@wraps_ufunc(
    return_units=dict(mean_charge_state=ureg.dimensionless),
    input_units=dict(
        average_electron_density=ureg.m**-3,
        average_electron_temp=ureg.eV,
        impurity_species=None,
        atomic_data=None,
    ),
    pass_as_kwargs=("atomic_data",),
)
def calc_impurity_charge_state(
    average_electron_density: float,
    average_electron_temp: float,
    impurity_species: Impurity,
    atomic_data: dict[Impurity, xr.DataArray],
) -> float:
    """Calculate the impurity charge state of the specified impurity species.

    Args:
        average_electron_density: [1e19 m^-3] :term:`glossary link<average_electron_density>`
        average_electron_temp: [keV] :term:`glossary link<average_electron_temp>`
        impurity_species: [] :term:`glossary link<impurity_species>`
        atomic_data: :term:`glossary link<atomic_data>`

    Returns:
        :term:`impurity_charge_state`
    """
    mean_charge_curve = atomic_data[impurity_species].coronal_mean_Z_interpolator
    return float(
        np.squeeze(np.power(10, mean_charge_curve(np.log10(average_electron_temp), np.log10(average_electron_density), grid=True)))
    )


def calc_change_in_zeff(impurity_charge_state: float, impurity_concentration: xr.DataArray) -> xr.DataArray:
    """Calculate the change in the effective charge due to the specified impurities.

    Args:
        impurity_charge_state: [~] :term:`glossary link<impurity_charge_state>`
        impurity_concentration: [~] :term:`glossary link<impurity_concentration>`

    Returns:
        change in zeff [~]
    """
    return impurity_charge_state * (impurity_charge_state - 1.0) * impurity_concentration


def calc_change_in_dilution(impurity_charge_state: float, impurity_concentration: xr.DataArray) -> xr.DataArray:
    """Calculate the change in n_fuel/n_e due to the specified impurities.

    Args:
        impurity_charge_state: [~] :term:`glossary link<impurity_charge_state>`
        impurity_concentration: [~] :term:`glossary link<impurity_concentration>`

    Returns:
        change in dilution [~]
    """
    return impurity_charge_state * impurity_concentration



"""Routines to calculate the upstream (unmitigated) parallel heat flux density."""
from numpy import pi

from ...unit_handling import Unitfull


def calc_parallel_heat_flux_density(
    power_crossing_separatrix: Unitfull,
    fraction_of_P_SOL_to_divertor: Unitfull,
    upstream_major_radius: Unitfull,
    lambda_q: Unitfull,
    upstream_fieldline_pitch: Unitfull,
) -> Unitfull:
    """Calculate the parallel heat flux density entering the flux tube (q_par).

    This expression is power to target divided by the area perpendicular to the flux tube.
    1. Power to target = power crossing separatrix * fraction of that power going to the target considered
    2. The poloidal area of a ring at the outboard midplane is 2 * pi * (R + minor_radius) * width
    3. For the width, we take the heat flux decay length lambda_q
    4. P_SOL * f_share / (2 * pi * (R + minor_radius) lambda_q) gives the heat flux per poloidal area
    5. We project this poloidal heat flux density into a parallel heat flux density by dividing by the field-line pitch

    Args:
      power_crossing_separatrix: [MW] :term:`glossary link<power_crossing_separatrix>`
      fraction_of_P_SOL_to_divertor: :term:`glossary link <fraction_of_P_SOL_to_divertor>`
      upstream_major_radius: [m] R + minor_radius, major radius at outboard midplane separatrix
      lambda_q: [mm] :term:`glossary link<lambda_q>`
      upstream_fieldline_pitch: B_total / B_poloidal at outboard midplane separatrix [~]

    Returns:
      q_parallel [GW/m^2]
    """
    return (
        power_crossing_separatrix * fraction_of_P_SOL_to_divertor / (2.0 * pi * upstream_major_radius * lambda_q) * upstream_fieldline_pitch
    )



"""Calculate toroidal, poloidal, total and normalized beta."""
from .. import formulas
from ..unit_handling import Unitfull, convert_to_default_units
from .algorithm_class import Algorithm

RETURN_KEYS = [
    "beta_toroidal",
    "beta_poloidal",
    "beta",
    "normalized_beta",
]


def run_calc_beta(
    average_electron_density: Unitfull,
    average_electron_temp: Unitfull,
    average_ion_temp: Unitfull,
    magnetic_field_on_axis: Unitfull,
    plasma_current: Unitfull,
    minor_radius: Unitfull,
) -> dict[str, Unitfull]:
    """Calculate toroidal, poloidal, total and normalized beta.

    Args:
        average_electron_density: :term:`glossary link<average_electron_density>`
        average_electron_temp: :term:`glossary link<average_electron_temp>`
        average_ion_temp: :term:`glossary link<average_ion_temp>`
        magnetic_field_on_axis: :term:`glossary link<magnetic_field_on_axis>`
        plasma_current: :term:`glossary link<plasma_current>`
        minor_radius: :term:`glossary link<minor_radius>`

    Returns:
        :term:`beta_toroidal`, :term:`beta_poloidal`, :term:`beta_total`, :term:`beta_N`
    """
    beta_toroidal = formulas.calc_beta_toroidal(average_electron_density, average_electron_temp, average_ion_temp, magnetic_field_on_axis)
    beta_poloidal = formulas.calc_beta_poloidal(
        average_electron_density, average_electron_temp, average_ion_temp, plasma_current, minor_radius
    )

    beta = formulas.calc_beta_total(beta_toroidal=beta_toroidal, beta_poloidal=beta_poloidal)
    normalized_beta = formulas.calc_beta_normalised(beta, minor_radius, magnetic_field_on_axis, plasma_current)

    local_vars = locals()
    return {key: convert_to_default_units(local_vars[key], key) for key in RETURN_KEYS}


calc_beta = Algorithm(
    function=run_calc_beta,
    return_keys=RETURN_KEYS,
)



"""Functions to reshape xarrays."""
from collections.abc import Sequence
from typing import Callable, Optional, Union

import numpy as np
import xarray as xr
from scipy.interpolate import griddata  # type:ignore[import]

from cfspopcon.unit_handling import Unit, magnitude


def order_dimensions(
    array: xr.DataArray,
    dims: Sequence[str],
    units: Optional[dict[str, Unit]] = None,
    template: Optional[Union[xr.DataArray, xr.Dataset]] = None,
    order_for_plotting: bool = True,
) -> xr.DataArray:
    """Reorder the dimensions of the array, broadcasting against `template` if necessary.

    This is particularly useful for plotting.
    """
    if template is None:
        template = xr.zeros_like(array)

    processed_dims = []
    for dim in dims:
        # We sometimes add a "dim_" in front of coord to differentiate between
        # variables and coordinates (in particular, coordinates can't have units)
        if dim not in array.dims and f"dim_{dim}" in array.dims:
            dim = f"dim_{dim}"  # noqa: PLW2901

        if dim in array.dims:
            processed_dims.append(dim)
        else:
            # If we've asked for a dimension, but didn't find it in array.dims,
            # look in the template.
            if dim not in template.dims and f"dim_{dim}" in template.dims:
                dim = f"dim_{dim}"  # noqa: PLW2901
            if dim in template.dims:
                # If we find the dimension in the template, broadcast the array
                # to have the correct dimensions.
                array = array.broadcast_like(template[dim])
                processed_dims.append(dim)
            else:
                raise ValueError(f"Array does not have dimension {dim.lstrip('dim_')}")

    if units is not None:
        for dim, processed_dim in zip(dims, processed_dims):
            array[processed_dim] = magnitude(template[dim.lstrip("dim_")].pint.to(units[dim]))

    if order_for_plotting:
        processed_dims = processed_dims[::-1]

    return array.transpose(*processed_dims)


def interpolate_array_onto_new_coords(
    array: xr.DataArray,
    new_coords: dict[str, xr.DataArray],
    resolution: Optional[dict[str, int]] = None,
    coord_min: Optional[dict[str, int]] = None,
    coord_max: Optional[dict[str, int]] = None,
    default_resolution: int = 50,
    max_distance: float = 2.0,
    griddata_method: str = "linear",
) -> xr.DataArray:
    """Take an xarray of values and map it to a grid of new coordinates.

    The input array and the new_coords must share the same coordinates (or be
    able to be broadcast to the same coordinates).

    The new mesh will usually not overlap perfectly with the original mesh, and
    may include regions not included in the original mesh. These regions are set
    to NaN (based on the distance between a sample point and a new-grid point).

    This method works for arbitrary number of new_coords, but works best and makes
    most sense for 2D (two new_coords).
    """
    coords, coord_spacing = dict(), dict()
    for key, coord_array in new_coords.items():
        coords[key], coord_spacing[key] = np.linspace(
            start=coord_min.get(key, magnitude(coord_array.min())) if coord_min else magnitude(coord_array.min()),
            stop=coord_max.get(key, magnitude(coord_array.max())) if coord_max else magnitude(coord_array.max()),
            num=resolution.get(key, default_resolution) if resolution else default_resolution,
            retstep=True,
        )

    mesh_coords = tuple(np.meshgrid(*coords.values(), indexing="ij"))

    broadcast_arrays = xr.broadcast(*(*list(new_coords.values()), array))
    sample_points = tuple([np.ravel(magnitude(broadcast_array)) for broadcast_array in broadcast_arrays[:-1]])
    array = broadcast_arrays[-1]

    interpolated_array = xr.DataArray(
        griddata(
            sample_points,
            np.ravel(magnitude(array)),
            mesh_coords,
            method=griddata_method,
            rescale=True,
        ),
        coords=coords,
        attrs=array.attrs,
    ).pint.quantify(array.pint.units)

    # Calculate the out-of-bounds mask based on distance to nearest sample point
    mesh_shape = [coord.size for coord in coords.values()]
    distance_to_nearest = np.zeros(mesh_shape, dtype=float)
    stacked_samples = np.vstack(sample_points)
    spacing = np.expand_dims(list(coord_spacing.values()), axis=-1)

    for index in np.ndindex(tuple(mesh_shape)):
        # Iterate over each grid point
        mesh_point = [mesh_coords[dimension][index] for dimension in range(len(index))]
        broadcast_mesh_point = np.broadcast_to(np.expand_dims(mesh_point, axis=-1), stacked_samples.shape)

        distance_to_points = ((broadcast_mesh_point - stacked_samples) / spacing) ** 2
        distance_to_nearest[index] = np.min(np.sum(distance_to_points, axis=0))

    clipped_array = interpolated_array.where(distance_to_nearest < max_distance).clip(min=array.min(), max=array.max())

    return clipped_array  # type:ignore[no-any-return]


def build_transform_function_from_dict(dataset: xr.Dataset, plot_params: dict) -> Callable[[xr.DataArray], xr.DataArray]:
    """Build a function which can be called on a field to return an array with transformed coordinates.

    The simplest function is a transpose function, which makes sure that the dimensions are correctly ordered
    for plotting.

    We can also build a more complicated function which transforms the field in terms of other computed fields,
    remapping the field onto new axes.
    """
    if "coords" in plot_params and "new_coords" in plot_params:
        raise ValueError("Can only pass one of 'coords' or 'new_coords'.")

    if "coords" in plot_params:
        xdim = plot_params["coords"]["x"]["dimension"]
        ydim = plot_params["coords"]["y"]["dimension"]
        units = {
            xdim: plot_params["coords"]["x"].get("units"),
            ydim: plot_params["coords"]["y"].get("units"),
        }

        return lambda array: order_dimensions(array, dims=(xdim, ydim), units=units, template=dataset, order_for_plotting=True)

    elif "new_coords" in plot_params:
        new_coords, new_coord_min, new_coord_max, new_coord_res = {}, {}, {}, {}

        for coord in ["y", "x"]:
            new_coord = plot_params["new_coords"][coord]
            new_key = new_coord["dimension"]
            field = dataset[new_key]
            new_coords[new_key] = field.pint.to(new_coord.get("units", field.pint.units))
            if "min" in new_coord:
                new_coord_min[new_key] = new_coord["min"]
            if "max" in new_coord:
                new_coord_max[new_key] = new_coord["max"]
            if "resolution" in new_coord:
                new_coord_res[new_key] = new_coord["resolution"]

        return lambda array: interpolate_array_onto_new_coords(
            array=array,
            new_coords=new_coords,
            resolution=new_coord_res,
            coord_min=new_coord_min,
            coord_max=new_coord_max,
            max_distance=plot_params["new_coords"].get("max_distance", 5.0),
        )

    else:
        raise NotImplementedError("Must provide either 'coords' or 'new_coords' for the transform function.")



"""Routines to calculate the heat flux decay length (lambda_q), for several different scalings."""

from ...named_options import LambdaQScaling
from ...unit_handling import ureg, wraps_ufunc


@wraps_ufunc(
    return_units=dict(lambda_q=ureg.millimeter),
    input_units=dict(
        lambda_q_scaling=None,
        average_total_pressure=ureg.atm,
        power_crossing_separatrix=ureg.megawatt,
        major_radius=ureg.meter,
        B_pol_omp=ureg.tesla,
        inverse_aspect_ratio=ureg.dimensionless,
    ),
)
def calc_lambda_q(
    lambda_q_scaling: LambdaQScaling,
    average_total_pressure: float,
    power_crossing_separatrix: float,
    major_radius: float,
    B_pol_omp: float,
    inverse_aspect_ratio: float,
) -> float:
    """Calculate SOL heat flux decay length (lambda_q) from a scaling.

    Args:
        lambda_q_scaling: :term:`glossary link<lambda_q_scaling>`
        average_total_pressure: [atm] :term:`glossary link <average_total_pressure>`
        power_crossing_separatrix: [MW] :term:`glossary link<power_crossing_separatrix>`
        major_radius: [m] :term:`glossary link<major_radius>`
        B_pol_omp: [T] :term:`glossary link<B_pol_omp>`
        inverse_aspect_ratio: [~] :term:`glossary link<inverse_aspect_ratio>`

    Returns:
        :term:`lambda_q` [mm]
    """
    if lambda_q_scaling == LambdaQScaling.Brunner:
        return float(calc_lambda_q_with_brunner.__wrapped__(average_total_pressure))
    elif lambda_q_scaling == LambdaQScaling.EichRegression14:
        return float(calc_lambda_q_with_eich_regression_14.__wrapped__(B_pol_omp))
    elif lambda_q_scaling == LambdaQScaling.EichRegression15:
        return float(
            calc_lambda_q_with_eich_regression_15.__wrapped__(power_crossing_separatrix, major_radius, B_pol_omp, inverse_aspect_ratio)
        )
    else:
        raise NotImplementedError(f"No implementation for lambda_q scaling {lambda_q_scaling}")


@wraps_ufunc(
    return_units=dict(lambda_q=ureg.millimeter),
    input_units=dict(average_total_pressure=ureg.atm),
)
def calc_lambda_q_with_brunner(average_total_pressure: float) -> float:
    """Return lambda_q according to the Brunner scaling.

    Equation 4 in :cite:`brunner_2018_heat_flux`
    """
    return float(0.91 * average_total_pressure**-0.48)


@wraps_ufunc(return_units=dict(lambda_q=ureg.millimeter), input_units=dict(B_pol_omp=ureg.tesla))
def calc_lambda_q_with_eich_regression_14(B_pol_omp: float) -> float:
    """Return lambda_q according to Eich regression 14.

    #14 in Table 3 in :cite:`eich_scaling_2013`
    """
    return float(0.63 * B_pol_omp**-1.19)


@wraps_ufunc(
    return_units=dict(lambda_q=ureg.millimeter),
    input_units=dict(
        power_crossing_separatrix=ureg.megawatt,
        major_radius=ureg.meter,
        B_pol_omp=ureg.tesla,
        inverse_aspect_ratio=ureg.dimensionless,
    ),
)
def calc_lambda_q_with_eich_regression_15(
    power_crossing_separatrix: float, major_radius: float, B_pol_omp: float, inverse_aspect_ratio: float
) -> float:
    """Return lambda_q according to Eich regression 15.

    #15 in Table 3 in :cite:`eich_scaling_2013`
    """
    lambda_q = 1.35 * major_radius**0.04 * B_pol_omp**-0.92 * inverse_aspect_ratio**0.42
    if power_crossing_separatrix > 0:
        return float(lambda_q * power_crossing_separatrix**-0.02)
    else:
        return float(lambda_q)



"""Calculate dependent geometry parameters."""
from .. import formulas
from ..unit_handling import Unitfull, convert_to_default_units
from .algorithm_class import Algorithm

RETURN_KEYS = [
    "separatrix_elongation",
    "separatrix_triangularity",
    "minor_radius",
    "vertical_minor_radius",
    "plasma_volume",
    "surface_area",
]


def run_calc_geometry(
    major_radius: Unitfull,
    inverse_aspect_ratio: Unitfull,
    areal_elongation: Unitfull,
    triangularity_psi95: Unitfull,
    elongation_ratio_sep_to_areal: Unitfull,
    triangularity_ratio_sep_to_psi95: Unitfull,
) -> dict[str, Unitfull]:
    """Calculate dependent geometry parameters.

    Args:
        major_radius: :term:`glossary link<major_radius>`
        inverse_aspect_ratio: :term:`glossary link<inverse_aspect_ratio>`
        areal_elongation: :term:`glossary link<areal_elongation>`
        triangularity_psi95: :term:`glossary link<triangularity_psi95>`
        elongation_ratio_sep_to_areal: :term:`glossary link<elongation_ratio_sep_to_areal>`
        triangularity_ratio_sep_to_psi95: :term:`glossary link<triangularity_ratio_sep_to_psi95>`

    Returns:
        :term:`separatrix_elongation`, :term:`separatrix_triangularity`, :term:`minor_radius`, :term:`vertical_minor_radius`, :term:`plasma_volume`, :term:`surface_area`
    """
    separatrix_elongation = areal_elongation * elongation_ratio_sep_to_areal

    separatrix_triangularity = triangularity_psi95 * triangularity_ratio_sep_to_psi95

    minor_radius = major_radius * inverse_aspect_ratio
    vertical_minor_radius = minor_radius * separatrix_elongation

    plasma_volume = formulas.calc_plasma_volume(major_radius, inverse_aspect_ratio, areal_elongation)
    surface_area = formulas.calc_plasma_surface_area(major_radius, inverse_aspect_ratio, areal_elongation)

    local_vars = locals()
    return {key: convert_to_default_units(local_vars[key], key) for key in RETURN_KEYS}


calc_geometry = Algorithm(
    function=run_calc_geometry,
    return_keys=RETURN_KEYS,
)



"""Plot creation functions."""
from pathlib import Path
from typing import Optional

import matplotlib
import matplotlib.contour
import matplotlib.pyplot as plt
import numpy as np
import xarray as xr
from matplotlib.axes import Axes

from cfspopcon import __version__

from ..point_selection import build_mask_from_dict, find_coords_of_minimum
from ..transform import build_transform_function_from_dict
from ..unit_handling import Quantity, Unit, dimensionless_magnitude
from .coordinate_formatter import CoordinateFormatter


def make_plot(
    dataset: xr.Dataset, plot_params: dict, points: dict, title: str, ax: Optional[Axes] = None, output_dir: Path = Path(".")
) -> None:
    """Given a dictionary corresponding to a plotting style, build a standard plot from the results of the POPCON."""
    if plot_params["type"] == "popcon":
        if ax is None:
            _, ax = plt.subplots(figsize=plot_params["figsize"], dpi=plot_params["show_dpi"])
        fig, ax = make_popcon_plot(dataset, title, plot_params, points, ax=ax)
    else:
        raise NotImplementedError(f"No plotting method for type '{plot_params['type']}'")

    if "save_as" in plot_params.keys() and output_dir is not None:
        fig.savefig(output_dir / plot_params["save_as"])


def make_popcon_plot(dataset: xr.Dataset, title: str, plot_params: dict, points: dict, ax: Axes):
    """Make a plot."""
    fig = ax.figure
    transform_func = build_transform_function_from_dict(dataset, plot_params)

    coords = plot_params["coords"] if "coords" in plot_params else plot_params["new_coords"]
    legend_elements = dict()

    if "fill" in plot_params:
        # Make a filled plot (max 1 variable)
        subplot_params = plot_params["fill"]
        field = dataset[subplot_params["variable"]]
        units = subplot_params.get("units", field.pint.units)
        field = field.pint.to(units)

        mask = build_mask_from_dict(dataset, plot_params=subplot_params)
        transformed_field = transform_func(field.where(mask))

        im = transformed_field.plot(ax=ax, add_colorbar=False)
        cbar = fig.colorbar(im, ax=ax)

        ax.format_coord = CoordinateFormatter(transformed_field)

        cbar.ax.set_ylabel(
            f"{subplot_params.get('cbar_label', subplot_params['variable'])} {units_to_string(field.pint.units)}",
            rotation=270,
            labelpad=subplot_params.get("labelpad", 15.0),
        )

    if "contour" in plot_params:
        # Overlay contour plots

        for variable, subplot_params in plot_params["contour"].items():

            field = dataset[variable]
            units = subplot_params.get("units", field.pint.units)
            field = field.pint.to(units)

            mask = build_mask_from_dict(dataset, plot_params=subplot_params)
            transformed_field = transform_func(field.where(mask))

            contour_set = transformed_field.plot.contour(
                ax=ax, levels=subplot_params["levels"], colors=[subplot_params["color"]], linestyles=[subplot_params.get("line", "solid")]
            )
            legend_entry = label_contour(
                ax, contour_set, fontsize=subplot_params.get("fontsize", 10.0), format_spec=subplot_params.get("format", "")
            )

            legend_elements[subplot_params["label"]] = legend_entry

    for key, point_params in points.items():
        point_style = plot_params.get("points", dict()).get(key, dict())
        label = point_style.get("label", key)

        mask = build_mask_from_dict(dataset, point_params)

        if "minimize" not in point_params.keys() and "maximize" not in point_params.keys():
            raise ValueError(f"Need to provide either minimize or maximize in point specification. Keys were {point_params.keys()}")

        if "minimize" in point_params.keys():
            field = dataset[point_params["minimize"]]
        else:
            field = -dataset[point_params["maximize"]]

        transformed_field = transform_func(field.where(mask))

        point_coords = find_coords_of_minimum(transformed_field, keep_dims=point_params.get("keep_dims", []))

        point = transformed_field.isel(point_coords)
        plotting_coords = []
        for dim in [coords["x"]["dimension"], coords["y"]["dimension"]]:
            if dim not in point.coords and f"dim_{dim}" in point.coords:
                dim = f"dim_{dim}"  # noqa: PLW2901
            plotting_coords.append(point[dim])

        legend_elements[label] = ax.scatter(
            *plotting_coords,
            s=point_style.get("size", None),
            c=point_style.get("color", None),
            marker=point_style.get("marker", None),
        )

    ax.set_title(f"{title} [{__version__}]")
    ax.set_xlabel(coords["x"]["label"])
    ax.set_ylabel(coords["y"]["label"])
    ax.legend(legend_elements.values(), legend_elements.keys())
    plt.tight_layout()

    return fig, ax


def units_to_string(units: Unit) -> str:
    """Given a pint Unit, return a string to represent the unit."""
    dummy_var = Quantity(1.0, units)
    if dummy_var.check("") and np.isclose(dimensionless_magnitude(dummy_var), 1.0):
        return ""
    else:
        return f"[{units:~P}]"


def label_contour(ax: plt.Axes, contour_set: matplotlib.contour.QuadContourSet, format_spec: str = "1.1f", fontsize: float = 10.0):
    """Add in-line labels to contours.

    Returns the first label element, which can be used to construct a legend.
    Works best with contour sets with only one color.
    >>> contour_labels = dict()
    >>> contour_labels["key"] = label_contour(...)
    >>> ax.legend(contour_labels.values(), contour_labels.keys())

    Inputs:
        ax: the matplotlib axis of the contour plot
        contour_set: the return argument of plt.contour(...)
        format_spec: a format specification for the string formatting
        fontsize: the font size of the label
    """

    def fmt(x):
        return f"{x:{format_spec}}"

    ax.clabel(contour_set, contour_set.levels, inline=True, fmt=fmt, fontsize=fontsize)

    return contour_set.legend_elements()[0][0]



"""Calculate tau_e and P_in from a tau_e scaling and the stored energy."""
from pathlib import Path

import numpy as np
import yaml

from ...named_options import ConfinementScaling
from ...unit_handling import ureg, wraps_ufunc

# Preload the scalings (instead of doing fileio in loop)
with open(Path(__file__).parent / "tau_e_scalings.yaml") as f:
    TAU_E_SCALINGS = yaml.safe_load(f)


@wraps_ufunc(
    return_units=dict(tau_e=ureg.s, P_tau=ureg.MW),
    input_units=dict(
        confinement_time_scalar=ureg.dimensionless,
        plasma_current=ureg.MA,
        magnetic_field_on_axis=ureg.T,
        average_electron_density=ureg.n19,
        major_radius=ureg.m,
        areal_elongation=ureg.dimensionless,
        separatrix_elongation=ureg.dimensionless,
        inverse_aspect_ratio=ureg.dimensionless,
        fuel_average_mass_number=ureg.amu,
        triangularity_psi95=ureg.dimensionless,
        separatrix_triangularity=ureg.dimensionless,
        plasma_stored_energy=ureg.MJ,
        q_star=ureg.dimensionless,
        tau_e_scaling=None,
    ),
    output_core_dims=[(), ()],
)
def calc_tau_e_and_P_in_from_scaling(
    confinement_time_scalar: float,
    plasma_current: float,
    magnetic_field_on_axis: float,
    average_electron_density: float,
    major_radius: float,
    areal_elongation: float,
    separatrix_elongation: float,
    inverse_aspect_ratio: float,
    fuel_average_mass_number: float,
    triangularity_psi95: float,
    separatrix_triangularity: float,
    plasma_stored_energy: float,
    q_star: float,
    tau_e_scaling: ConfinementScaling,
) -> tuple[float, float]:
    r"""Calculate energy confinement time and input power from a tau_E scaling.

    The energy confinement time can generally be written as

    .. math::
        \tau_e = H \cdot C \cdot P_{\tau}^{\alpha_P}
        \cdot I_{MA}^{\alpha_I} \cdot B_0^{\alpha_B} \cdot \bar{n_{e,19}}^{\alpha_n} \cdot R_0^{\alpha_R}
        \cdot \kappa_A^{\alpha_{ka}} \cdot \kappa_{sep}^{\alpha_{ks}} \cdot \epsilon^{\alpha_e}
        \cdot m_i^{\alpha_A} \cdot \delta^{\alpha_d}

    We don't know :math:`P_{\tau}` in advance, so instead write

    .. math:: \tau_e = \gamma \cdot P_{\tau}^{\alpha_P}

    with

    .. math::
        \gamma = H \cdot C
        \cdot I_{MA}^{\alpha_I} \cdot B_0^{\alpha_B} \cdot \bar{n_{e,19}}^{\alpha_n} \cdot R_0^{\alpha_R}
        \cdot \kappa_A^{\alpha_{ka}} \cdot \kappa_{sep}^{\alpha_{ks}} \cdot \epsilon^{\alpha_e}
        \cdot m_i^{\alpha_A} \cdot \delta^{\alpha_d}

    We have everything that we need to evaluate :math:`\gamma`. Then, we also know that

    .. math:: \tau_e = W_p / P_{loss}

    Then, we crucially need to define what exactly we mean by the two powers that we've introduced
    (:math:`P_{\tau}` and :math:`P_{loss}`). We usually take
    :math:`P_{\tau} = P_{heating} = P_{ohmic} + P_{\alpha} + P_{aux}` and
    :math:`P_{loss}=P_{SOL} + P_{rad,core}` [Wesson definition]. From a
    simple power balance, :math:`P_{heating}=P_{loss}` and so, setting the two :math:`\tau_e` equations equal, we get that

    .. math::
        W_p / P = \gamma \cdot P^{\alpha_P}
        P^{\alpha_P + 1} = W_p / \gamma
        P = \left(W_p / \gamma \right)^{\frac{1}{\alpha_P + 1}}

    Once we have :math:`P = P_{ohmic} + P_{\alpha} + P_{aux} = P_{SOL} + P_{rad,core}`, we can calculate :math:`W_p/P`.

    However, it is also possible that the core radiated power is subtracted when calculating :math:`\tau_e`
    [Freidberg definition of :math:`W_p`], giving

    .. math::
        P_{\tau} = P_{ohmic} + P_{\alpha} + P_{aux} - P_{rad} = P_{loss} = P_{SOL}

    If you are using a scaling where this is the case, set ``tau_e_scaling_uses_P_in=False``.
    Then, the returned value should be interpreted as :math:`P_{SOL}`.

    N.b. there are two more possible cases, where different powers are used in the two :math:`\tau_e` scalings.
    We don't allow these cases, since 1) experiments generally pick a consistent definition for :math:`P`
    and 2) this results in an equation for :math:`P` which cannot be solved analytically.

    Args:
        confinement_time_scalar: [~] confinement scaling factor
        plasma_current: [MA] :term:`glossary link<plasma_current>`
        magnetic_field_on_axis: [T] :term:`glossary link<magnetic_field_on_axis>`
        average_electron_density: [1e19 m^-3] :term:`glossary link<average_electron_density>`
        major_radius: [m] :term:`glossary link<major_radius>`
        areal_elongation: [~] :term:`glossary link<areal_elongation>`
        separatrix_elongation: [~] :term:`glossary link<separatrix_elongation>`
        inverse_aspect_ratio: [~] :term:`glossary link<inverse_aspect_ratio>`
        fuel_average_mass_number: [amu] :term:`glossary link<fuel_average_mass_number>`
        triangularity_psi95: [~] :term:`glossary link<triangularity_psi95>`
        separatrix_triangularity: [~] :term:`glossary link<separatrix_triangularity>`
        plasma_stored_energy: [MJ] :term:`glossary link<plasma_stored_energy>`
        q_star: [~] :term:`glossary link<q_star>`
        tau_e_scaling: [] :term:`glossary link<tau_e_scaling>`

    Returns:
        :term:`energy_confinement_time` [s], :term:`P_in` if tau_e_scaling_uses_P_in=False, else :term:`P_SOL` [MW]
    """
    scaling = TAU_E_SCALINGS[tau_e_scaling.name]

    gamma = (
        confinement_time_scalar
        * scaling["params"]["C"]
        * plasma_current ** scaling["params"]["a_I"]
        * magnetic_field_on_axis ** scaling["params"]["a_B"]
        * average_electron_density ** scaling["params"]["a_n"]
        * major_radius ** scaling["params"]["a_R"]
        * areal_elongation ** scaling["params"]["a_ka"]
        * separatrix_elongation ** scaling["params"]["a_ks"]
        * inverse_aspect_ratio ** scaling["params"]["a_e"]
        * fuel_average_mass_number ** scaling["params"]["a_A"]
        * (1.0 + np.mean([triangularity_psi95, separatrix_triangularity])) ** scaling["params"]["a_d"]
        * q_star ** scaling["params"]["a_q"]
    )

    if gamma > 0.0:
        P_tau = (plasma_stored_energy / gamma) ** (1.0 / (1.0 + scaling["params"]["a_P"]))
    else:
        P_tau = np.inf

    tau_E = plasma_stored_energy / P_tau

    return float(tau_E), float(P_tau)


def load_metadata(dataset: str) -> dict[str, str]:
    """Load dataset metadata from YAML file.

    Args:
        dataset: name of scaling in ./energy_confinement_time.yaml

    Returns:
        Metadata
    """
    metadata: dict[str, str] = TAU_E_SCALINGS[dataset]["metadata"]
    return metadata


def get_datasets() -> list[str]:
    """Get a list of names of valid datasets.

    Returns:
        List of names of valid datasets
    """
    datasets: list[str] = list(TAU_E_SCALINGS.keys())  # Unpack iterator to list
    return datasets



"""Calculate the impurity radiated power using the radas atomic_data."""
import numpy as np
import xarray as xr
from numpy import float64
from numpy.typing import NDArray

from ...named_options import Impurity
from ...unit_handling import magnitude, ureg, wraps_ufunc
from ..helpers import integrate_profile_over_volume


@wraps_ufunc(
    return_units=dict(radiated_power=ureg.MW),
    input_units=dict(
        rho=ureg.dimensionless,
        electron_temp_profile=ureg.eV,
        electron_density_profile=ureg.m**-3,
        impurity_concentration=ureg.dimensionless,
        impurity_species=None,
        plasma_volume=ureg.m**3,
        atomic_data=None,
    ),
    input_core_dims=[("dim_rho",), ("dim_rho",), ("dim_rho",), (), (), ()],
    pass_as_kwargs=("atomic_data",),
)
def calc_impurity_radiated_power_radas(
    rho: NDArray[float64],
    electron_temp_profile: NDArray[float64],
    electron_density_profile: NDArray[float64],
    impurity_concentration: float,
    impurity_species: Impurity,
    plasma_volume: float,
    atomic_data: dict[Impurity, xr.DataArray],
) -> float:
    """Calculation of radiated power using radas atomic_data datasets.

    Args:
        rho: [~] :term:`glossary link<rho>`
        electron_temp_profile: [keV] :term:`glossary link<electron_temp_profile>`
        electron_density_profile: [1e19 m^-3] :term:`glossary link<electron_density_profile>`
        impurity_species: [] :term:`glossary link<impurity_species>`
        impurity_concentration: [~] :term:`glossary link<impurity_concentration>`
        plasma_volume: [m^3] :term:`glossary link<plasma_volume>`
        atomic_data: :term:`glossary link<atomic_data>`

    Returns:
         [MW] Estimated radiation power due to this impurity
    """
    MW_per_W = 1e6

    ds = atomic_data[impurity_species]
    Lz_curve = ds.coronal_Lz_interpolator

    # Use nearest neighbor extrapolation if evaluating for a
    # point off-grid

    electron_temp_profile = np.minimum(electron_temp_profile, magnitude(ds.electron_temperature.max()))
    electron_temp_profile = np.maximum(electron_temp_profile, magnitude(ds.electron_temperature.min()))
    electron_density_profile = np.minimum(electron_density_profile, magnitude(ds.electron_density.max()))
    electron_density_profile = np.maximum(electron_density_profile, magnitude(ds.electron_density.min()))

    Lz = np.power(10, Lz_curve(np.log10(electron_temp_profile), np.log10(electron_density_profile), grid=False))
    radiated_power_profile = electron_density_profile**2 * Lz

    radiated_power = impurity_concentration * integrate_profile_over_volume(radiated_power_profile, rho, plasma_volume) / MW_per_W

    return radiated_power



"""Compute the total radiated power."""
import numpy as np
import xarray as xr

from ...named_options import Impurity, RadiationMethod
from ...unit_handling import Unitfull, ureg
from .mavrin_coronal import calc_impurity_radiated_power_mavrin_coronal
from .mavrin_noncoronal import calc_impurity_radiated_power_mavrin_noncoronal
from .post_and_jensen import calc_impurity_radiated_power_post_and_jensen
from .radas import calc_impurity_radiated_power_radas


def calc_impurity_radiated_power(
    radiated_power_method: RadiationMethod,
    rho: Unitfull,
    electron_temp_profile: Unitfull,
    electron_density_profile: Unitfull,
    impurities: xr.DataArray,
    plasma_volume: Unitfull,
    atomic_data: dict[Impurity, xr.DataArray],
) -> xr.DataArray:
    """Compute the total radiated power due to fuel and impurity species.

    Args:
        radiated_power_method: [] :term:`glossary link<radiated_power_method>`
        rho: [~] :term:`glossary link<rho>`
        electron_temp_profile: [keV] :term:`glossary link<electron_temp_profile>`
        electron_density_profile: [1e19 m^-3] :term:`glossary link<electron_density_profile>`
        impurities: [] :term:`glossary link<impurities>`
        plasma_volume: [m^3] :term:`glossary link<plasma_volume>`
        atomic_data: :term:`glossary link<atomic_data>`

    Returns:
         [MW] Estimated radiation power due to this impurity
    """
    P_rad_kwargs = dict(
        rho=rho,
        electron_temp_profile=electron_temp_profile,
        electron_density_profile=electron_density_profile,
        impurity_concentration=impurities,
        impurity_species=impurities.dim_species,
        plasma_volume=plasma_volume,
    )
    if radiated_power_method == RadiationMethod.PostJensen:
        P_rad_impurity = calc_impurity_radiated_power_post_and_jensen(**P_rad_kwargs)
    elif radiated_power_method == RadiationMethod.MavrinCoronal:
        P_rad_impurity = calc_impurity_radiated_power_mavrin_coronal(**P_rad_kwargs)
    elif radiated_power_method == RadiationMethod.MavrinNoncoronal:
        P_rad_impurity = calc_impurity_radiated_power_mavrin_noncoronal(**P_rad_kwargs, tau_i=np.inf * ureg.s)
    elif radiated_power_method == RadiationMethod.Radas:
        P_rad_impurity = calc_impurity_radiated_power_radas(**P_rad_kwargs, atomic_data=atomic_data)
    else:
        raise NotImplementedError(f"No implementation for radiated_power_method = {radiated_power_method}")

    return P_rad_impurity  # type:ignore[no-any-return]



"""Define default units for writing to/from disk."""
from collections.abc import Iterable
from numbers import Number
from typing import Any, Union, overload

import numpy as np
import xarray as xr

from .setup_unit_handling import DimensionalityError, Quantity, convert_units, magnitude

DEFAULT_UNITS = dict(
    areal_elongation="",
    average_electron_density="n19",
    average_electron_temp="keV",
    average_ion_density="n19",
    average_ion_temp="keV",
    average_total_pressure="Pa",
    B_pol_out_mid="T",
    B_t_out_mid="T",
    beta_poloidal="",
    beta_toroidal="",
    beta="",
    bootstrap_fraction="",
    confinement_threshold_scalar="",
    confinement_time_scalar="",
    core_radiated_power_fraction="",
    core_radiator_charge_state="",
    core_radiator_concentration="",
    core_radiator=None,
    current_relaxation_time="s",
    dilution_change_from_core_rad="",
    dilution="",
    effective_collisionality="",
    electron_density_peaking_offset="",
    electron_density_peaking="",
    electron_density_profile="n19",
    electron_temp_profile="keV",
    elongation_ratio_sep_to_areal="",
    energy_confinement_scaling=None,
    energy_confinement_time="s",
    f_shaping="",
    fieldline_pitch_at_omp="",
    fraction_of_external_power_coupled="",
    fraction_of_P_SOL_to_divertor="",
    fuel_average_mass_number="amu",
    fuel_ion_density_profile="n19",
    fusion_reaction=None,
    fusion_triple_product="n20 * keV * s",
    greenwald_fraction="",
    heavier_fuel_species_fraction="",
    impurities="",
    impurity_charge_state="",
    input_SOL_power_loss_fraction="",
    input_target_electron_temp="eV",
    input_target_q_parallel="GW / m**2",
    inverse_aspect_ratio="",
    ion_density_peaking_offset="",
    ion_density_peaking="",
    ion_temp_profile="keV",
    ion_to_electron_temp_ratio="",
    kappa_e0="W / (eV**3.5 m)",
    lambda_q_factor="",
    lambda_q_scaling=None,
    lambda_q="mm",
    loop_voltage="V",
    magnetic_field_on_axis="T",
    major_radius="m",
    minimum_core_radiated_fraction="",
    minor_radius="m",
    neoclassical_loop_resistivity="m * ohm",
    nesep_over_nebar="",
    neutron_power_flux_to_walls="MW / m**2",
    neutron_rate="s**-1",
    normalized_beta="percent * m * T / MA",
    normalized_inverse_temp_scale_length="",
    nu_star="",
    P_alpha="MW",
    P_auxillary="MW",
    P_external="MW",
    P_fusion="MW",
    P_in="MW",
    P_launched="MW",
    P_LH_thresh="MW",
    P_neutron="MW",
    P_ohmic="MW",
    P_radiated_by_core_radiator="MW",
    P_radiation="MW",
    P_sol="MW",
    parallel_connection_length="m",
    PB_over_R="MW * T / m",
    PBpRnSq="MW * T / m * n20**-2",
    peak_electron_density="n19",
    peak_electron_temp="keV",
    peak_fuel_ion_density="n19",
    peak_ion_temp="keV",
    peak_pressure="Pa",
    plasma_current="A",
    plasma_stored_energy="MJ",
    plasma_volume="m**3",
    product_of_magnetic_field_and_radius="m * T",
    profile_form=None,
    q_parallel="GW / m**2",
    q_perp="MW / m**2",
    q_star="",
    Q="",
    radiated_power_method=None,
    radiated_power_scalar="",
    ratio_of_P_SOL_to_P_LH="",
    rho_star="",
    rho="",
    separatrix_elongation="",
    separatrix_triangularity="",
    SOC_LOC_ratio="",
    SOL_momentum_loss_function=None,
    SOL_power_loss_fraction="",
    spitzer_resistivity="m * ohm",
    summed_impurity_density="n19",
    surface_area="m**2",
    target_electron_density="n19",
    target_electron_flux="m**-2 s**-1",
    target_electron_temp="eV",
    target_q_parallel="GW / m**2",
    tau_e_scaling_uses_P_in=None,
    temperature_peaking="",
    toroidal_flux_expansion="",
    trapped_particle_fraction="",
    triangularity_psi95="",
    triangularity_ratio_sep_to_psi95="",
    two_point_model_method=None,
    upstream_electron_temp="eV",
    vertical_minor_radius="m",
    z_effective="",
    zeff_change_from_core_rad="",
)


def default_unit(var: str) -> Union[str, None]:
    """Return cfspopcon's default unit for a given quantity.

    Args:
        var: Quantity name

    Returns: Unit
    """
    try:
        return DEFAULT_UNITS[var]
    except KeyError:
        raise KeyError(
            f"No default unit defined for {var}. Please check configured default units in the unit_handling submodule."
        ) from None


def magnitude_in_default_units(value: Union[Quantity, xr.DataArray], key: str) -> Union[float, list[float], Any]:
    """Convert values to default units and then return the magnitude.

    Args:
        value: input value to convert to a float
        key: name of field for looking up in DEFAULT_UNITS dictionary

    Returns:
        magnitude of value in default units and as basic type
    """
    try:
        # unit conversion step
        unit = default_unit(key)
        if unit is None:
            return value

        mag = magnitude(convert_units(value, unit))

    except DimensionalityError as e:
        print(f"Unit conversion failed for {key}. Could not convert '{value}' to '{DEFAULT_UNITS[key]}'")
        raise e

    # single value arrays -> float
    # np,xr array -> list
    if isinstance(mag, (np.ndarray, xr.DataArray)):
        if mag.size == 1:
            return float(mag)
        else:
            return [float(v) for v in mag]
    else:
        return float(mag)


@overload
def set_default_units(value: Number, key: str) -> Quantity:
    ...


@overload
def set_default_units(value: xr.DataArray, key: str) -> xr.DataArray:
    ...


@overload
def set_default_units(value: Any, key: str) -> Any:
    ...


def set_default_units(value: Any, key: str) -> Any:
    """Return value as a quantity with default units.

    Args:
        value: magnitude of input value to convert to a Quantity
        key: name of field for looking up in DEFAULT_UNITS dictionary

    Returns:
        magnitude of value in default units
    """

    def _is_number_not_bool(val: Any) -> bool:
        return isinstance(val, Number) and not isinstance(val, bool)

    def _is_iterable_of_number_not_bool(val: Any) -> bool:
        if not isinstance(val, Iterable):
            return False

        if isinstance(val, (np.ndarray, xr.DataArray)) and val.ndim == 0:
            return _is_number_not_bool(val.item())

        return all(_is_number_not_bool(v) for v in value)

    # None is used to ignore class types
    if DEFAULT_UNITS[key] is None:
        if _is_number_not_bool(value) or _is_iterable_of_number_not_bool(value):
            raise RuntimeError(
                f"set_default_units for key {key} and value {value} of type {type(value)}: numeric types should carry units!"
            )
        return value
    elif isinstance(value, xr.DataArray):
        return value.pint.quantify(DEFAULT_UNITS[key])
    else:
        return Quantity(value, DEFAULT_UNITS[key])


@overload
def convert_to_default_units(value: float, key: str) -> float:
    ...


@overload
def convert_to_default_units(value: xr.DataArray, key: str) -> xr.DataArray:
    ...


@overload
def convert_to_default_units(value: Quantity, key: str) -> Quantity:
    ...


def convert_to_default_units(value: Union[float, Quantity, xr.DataArray], key: str) -> Union[float, Quantity, xr.DataArray]:
    """Convert an array or scalar to default units."""
    unit = DEFAULT_UNITS[key]
    if unit is None or unit == "":
        return value
    elif isinstance(value, (xr.DataArray, Quantity)):
        return convert_units(value, unit)
    else:
        raise NotImplementedError(f"No implementation for 'convert_to_default_units' with an array of type {type(value)} ({value})")



"""Calculate plasma current from edge safety factor."""
from .. import formulas
from ..unit_handling import Unitfull, convert_to_default_units
from .algorithm_class import Algorithm

RETURN_KEYS = [
    "f_shaping",
    "q_star",
]


def run_calc_q_star_from_plasma_current(
    magnetic_field_on_axis: Unitfull,
    major_radius: Unitfull,
    plasma_current: Unitfull,
    inverse_aspect_ratio: Unitfull,
    areal_elongation: Unitfull,
    triangularity_psi95: Unitfull,
) -> dict[str, Unitfull]:
    """Calculate plasma current from edge safety factor.

    Args:
        magnetic_field_on_axis: :term:`glossary link<magnetic_field_on_axis>`
        major_radius: :term:`glossary link<major_radius>`
        plasma_current: :term:`glossary link<plasma_current>`
        inverse_aspect_ratio: :term:`glossary link<inverse_aspect_ratio>`
        areal_elongation: :term:`glossary link<areal_elongation>`
        triangularity_psi95: :term:`glossary link<triangularity_psi95>`

    Returns:
        :term:`f_shaping`, :term:`q_star`,
    """
    f_shaping = formulas.calc_f_shaping(inverse_aspect_ratio, areal_elongation, triangularity_psi95)
    q_star = formulas.calc_q_star(magnetic_field_on_axis, major_radius, inverse_aspect_ratio, plasma_current, f_shaping)

    local_vars = locals()
    return {key: convert_to_default_units(local_vars[key], key) for key in RETURN_KEYS}


calc_q_star_from_plasma_current = Algorithm(
    function=run_calc_q_star_from_plasma_current,
    return_keys=RETURN_KEYS,
)



"""Formulas for POPCONs analysis."""

from . import energy_confinement_time_scalings, fusion_reaction_data, plasma_profile_data, radiated_power
from .average_fuel_ion_mass import calc_fuel_average_mass_number
from .beta import calc_beta_normalised, calc_beta_poloidal, calc_beta_toroidal, calc_beta_total
from .confinement_regime_threshold_powers import (
    calc_confinement_transition_threshold_power,
    calc_LH_transition_threshold_power,
    calc_LI_transition_threshold_power,
)
from .current_drive import (
    calc_bootstrap_fraction,
    calc_current_relaxation_time,
    calc_f_shaping,
    calc_loop_voltage,
    calc_neoclassical_loop_resistivity,
    calc_ohmic_power,
    calc_plasma_current,
    calc_q_star,
    calc_resistivity_trapped_enhancement,
    calc_Spitzer_loop_resistivity,
)
from .density_peaking import calc_density_peaking, calc_effective_collisionality
from .divertor_metrics import calc_B_pol_omp, calc_B_tor_omp
from .energy_confinement_time_scalings import calc_tau_e_and_P_in_from_scaling
from .figures_of_merit import calc_normalised_collisionality, calc_peak_pressure, calc_rho_star, calc_triple_product
from .fusion_rates import calc_fusion_power, calc_neutron_flux_to_walls
from .geometry import calc_plasma_surface_area, calc_plasma_volume
from .impurity_effects import calc_change_in_dilution, calc_change_in_zeff, calc_impurity_charge_state
from .operational_limits import calc_greenwald_density_limit, calc_greenwald_fraction, calc_troyon_limit
from .plasma_profiles import calc_1D_plasma_profiles
from .Q_thermal_gain_factor import thermal_calc_gain_factor
from .radiated_power import (
    calc_bremsstrahlung_radiation,
    calc_impurity_radiated_power,
    calc_impurity_radiated_power_mavrin_coronal,
    calc_impurity_radiated_power_mavrin_noncoronal,
    calc_impurity_radiated_power_post_and_jensen,
    calc_impurity_radiated_power_radas,
    calc_synchrotron_radiation,
)

__all__ = [
    "energy_confinement_time_scalings",
    "fusion_reaction_data",
    "calc_density_peaking",
    "calc_effective_collisionality",
    "plasma_profile_data",
    "calc_fuel_average_mass_number",
    "calc_B_pol_omp",
    "calc_B_tor_omp",
    "calc_beta_normalised",
    "calc_beta_poloidal",
    "calc_beta_toroidal",
    "calc_troyon_limit",
    "calc_greenwald_density_limit",
    "calc_beta_total",
    "calc_bootstrap_fraction",
    "calc_current_relaxation_time",
    "calc_f_shaping",
    "calc_fusion_power",
    "calc_greenwald_fraction",
    "calc_LH_transition_threshold_power",
    "calc_LI_transition_threshold_power",
    "calc_loop_voltage",
    "calc_resistivity_trapped_enhancement",
    "calc_neoclassical_loop_resistivity",
    "calc_neutron_flux_to_walls",
    "calc_normalised_collisionality",
    "calc_1D_plasma_profiles",
    "calc_ohmic_power",
    "calc_peak_pressure",
    "calc_plasma_current",
    "calc_plasma_surface_area",
    "calc_plasma_volume",
    "calc_rho_star",
    "calc_Spitzer_loop_resistivity",
    "calc_q_star",
    "calc_triple_product",
    "calc_tau_e_and_P_in_from_scaling",
    "thermal_calc_gain_factor",
    "calc_confinement_transition_threshold_power",
    "calc_change_in_zeff",
    "calc_change_in_dilution",
    "calc_bremsstrahlung_radiation",
    "calc_synchrotron_radiation",
    "calc_impurity_radiated_power_post_and_jensen",
    "calc_impurity_radiated_power_radas",
    "calc_impurity_radiated_power",
    "calc_impurity_radiated_power_mavrin_coronal",
    "calc_impurity_radiated_power_mavrin_noncoronal",
    "radiated_power",
    "calc_impurity_charge_state",
]



"""Calculate density peaking and the corresponding density and temperature profiles."""
from .. import formulas
from ..named_options import ProfileForm
from ..unit_handling import Unitfull, convert_to_default_units
from .algorithm_class import Algorithm

RETURN_KEYS = [
    "effective_collisionality",
    "ion_density_peaking",
    "electron_density_peaking",
    "peak_electron_density",
    "peak_fuel_ion_density",
    "peak_electron_temp",
    "peak_ion_temp",
    "rho",
    "electron_density_profile",
    "fuel_ion_density_profile",
    "electron_temp_profile",
    "ion_temp_profile",
]


def run_calc_peaked_profiles(
    profile_form: ProfileForm,
    average_electron_density: Unitfull,
    average_electron_temp: Unitfull,
    average_ion_temp: Unitfull,
    ion_density_peaking_offset: Unitfull,
    electron_density_peaking_offset: Unitfull,
    temperature_peaking: Unitfull,
    major_radius: Unitfull,
    z_effective: Unitfull,
    dilution: Unitfull,
    beta_toroidal: Unitfull,
    normalized_inverse_temp_scale_length: Unitfull,
) -> dict[str, Unitfull]:
    """Calculate density peaking and the corresponding density and temperature profiles.

    Args:
        profile_form: :term:`glossary link<profile_form>`
        average_electron_density: :term:`glossary link<average_electron_density>`
        average_electron_temp: :term:`glossary link<average_electron_temp>`
        average_ion_temp: :term:`glossary link<average_ion_temp>`
        ion_density_peaking_offset: :term:`glossary link<ion_density_peaking_offset>`
        electron_density_peaking_offset: :term:`glossary link<electron_density_peaking_offset>`
        temperature_peaking: :term:`glossary link<temperature_peaking>`
        major_radius: :term:`glossary link<major_radius>`
        z_effective: :term:`glossary link<z_effective>`
        dilution: :term:`glossary link<dilution>`
        beta_toroidal: :term:`glossary link<beta_toroidal>`
        normalized_inverse_temp_scale_length: :term:`glossary link<normalized_inverse_temp_scale_length>`

    Returns:
    `effective_collisionality`, :term:`ion_density_peaking`, :term:`electron_density_peaking`, :term:`peak_electron_density`, :term:`peak_electron_temp`, :term:`peak_ion_temp`, :term:`rho`, :term:`electron_density_profile`, :term:`fuel_ion_density_profile`, :term:`electron_temp_profile`, :term:`ion_temp_profile`

    """
    effective_collisionality = formulas.calc_effective_collisionality(
        average_electron_density, average_electron_temp, major_radius, z_effective
    )
    ion_density_peaking = formulas.calc_density_peaking(effective_collisionality, beta_toroidal, nu_noffset=ion_density_peaking_offset)
    electron_density_peaking = formulas.calc_density_peaking(
        effective_collisionality, beta_toroidal, nu_noffset=electron_density_peaking_offset
    )

    peak_electron_density = average_electron_density * electron_density_peaking
    peak_fuel_ion_density = average_electron_density * dilution * ion_density_peaking
    peak_electron_temp = average_electron_temp * temperature_peaking
    peak_ion_temp = average_ion_temp * temperature_peaking

    # Calculate the total fusion power by estimating density and temperature profiles and
    # using this to calculate fusion power profiles.
    (rho, electron_density_profile, fuel_ion_density_profile, electron_temp_profile, ion_temp_profile,) = formulas.calc_1D_plasma_profiles(
        profile_form,
        average_electron_density,
        average_electron_temp,
        average_ion_temp,
        electron_density_peaking,
        ion_density_peaking,
        temperature_peaking,
        dilution,
        normalized_inverse_temp_scale_length,
    )

    local_vars = locals()
    return {key: convert_to_default_units(local_vars[key], key) for key in RETURN_KEYS}


calc_peaked_profiles = Algorithm(
    function=run_calc_peaked_profiles,
    return_keys=RETURN_KEYS,
)



"""Run the two point model with a fixed power loss fraction in the SOL."""
from typing import Union

import xarray as xr

from ..formulas.scrape_off_layer_model import solve_two_point_model
from ..named_options import MomentumLossFunction
from ..unit_handling import Quantity, convert_to_default_units
from .algorithm_class import Algorithm

RETURN_KEYS = [
    "upstream_electron_temp",
    "target_electron_density",
    "target_electron_temp",
    "target_electron_flux",
    "target_q_parallel",
]


def run_two_point_model_fixed_fpow(
    SOL_power_loss_fraction: Union[float, xr.DataArray],
    q_parallel: Union[Quantity, xr.DataArray],
    parallel_connection_length: Union[Quantity, xr.DataArray],
    average_electron_density: Union[Quantity, xr.DataArray],
    nesep_over_nebar: Union[float, xr.DataArray],
    toroidal_flux_expansion: Union[float, xr.DataArray],
    fuel_average_mass_number: Union[Quantity, xr.DataArray],
    kappa_e0: Union[Quantity, xr.DataArray],
    SOL_momentum_loss_function: Union[MomentumLossFunction, xr.DataArray],
    raise_error_if_not_converged: bool = False,
) -> dict[str, Union[float, Quantity, xr.DataArray]]:
    """Run the two point model with a fixed power loss fraction in the SOL.

    Args:
        SOL_power_loss_fraction: :term:`glossary link<SOL_power_loss_fraction>`
        q_parallel: :term:`glossary link<q_parallel>`
        parallel_connection_length: :term:`glossary link<parallel_connection_length>`
        average_electron_density: :term:`glossary link<average_electron_density>`
        nesep_over_nebar: :term:`glossary link<nesep_over_nebar>`
        toroidal_flux_expansion: :term:`glossary link<toroidal_flux_expansion>`
        fuel_average_mass_number: :term:`glossary link<fuel_average_mass_number>`
        kappa_e0: :term:`glossary link<kappa_e0>`
        SOL_momentum_loss_function: :term:`glossary link<SOL_momentum_loss_function>`
        raise_error_if_not_converged: Raise an error if solve does not converge

    Returns:
        :term:`upstream_electron_temp`, :term:`target_electron_density`, :term:`target_electron_temp`, :term:`target_electron_flux`, :term:`target_q_parallel`,
    """
    (upstream_electron_temp, target_electron_density, target_electron_temp, target_electron_flux,) = solve_two_point_model(
        SOL_power_loss_fraction=SOL_power_loss_fraction,
        parallel_heat_flux_density=q_parallel,
        parallel_connection_length=parallel_connection_length,
        upstream_electron_density=nesep_over_nebar * average_electron_density,
        toroidal_flux_expansion=toroidal_flux_expansion,
        fuel_average_mass_number=fuel_average_mass_number,
        kappa_e0=kappa_e0,
        SOL_momentum_loss_function=SOL_momentum_loss_function,
        raise_error_if_not_converged=raise_error_if_not_converged,
    )

    target_q_parallel = q_parallel * (1.0 - SOL_power_loss_fraction)

    local_vars = locals()
    return {key: convert_to_default_units(local_vars[key], key) for key in RETURN_KEYS}


two_point_model_fixed_fpow = Algorithm(
    function=run_two_point_model_fixed_fpow,
    return_keys=RETURN_KEYS,
)



"""Defines a class for different POPCON algorithms."""
from __future__ import annotations

import inspect
from collections.abc import Callable, Sequence
from functools import wraps
from typing import Any, Optional, Union
from warnings import warn

import xarray as xr

from ..unit_handling import convert_to_default_units

FunctionType = Callable[..., dict[str, Any]]


class Algorithm:
    """A class which handles the input and output of POPCON algorithms."""

    def __init__(self, function: FunctionType, return_keys: list[str], name: Optional[str] = None):
        """Initialise an Algorithm.

        Args:
            function: a callable function
            return_keys: the arguments which are returned from the function
            name: Descriptive name for algorithm
        """
        self._function = function
        self._name = self._function.__name__ if name is None else name

        self._signature = inspect.signature(function)
        for p in self._signature.parameters.values():
            if p.kind not in (
                inspect.Parameter.POSITIONAL_OR_KEYWORD,
                inspect.Parameter.KEYWORD_ONLY,
                inspect.Parameter.VAR_KEYWORD,
            ):
                raise ValueError(
                    f"Algorithm only supports functions with keyword arguments, but {function}, has {p.kind} parameter {p.name}"
                )
        self.input_keys = list(self._signature.parameters.keys())
        self.return_keys = return_keys

        self.default_values = {
            key: val.default for key, val in self._signature.parameters.items() if val.default is not inspect.Parameter.empty
        }
        self.default_keys = list(self.default_values.keys())

        self.required_input_keys = [key for key in self.input_keys if key not in self.default_keys]

        self.__doc__ = self._make_docstring()

        self.run = self._make_run(self._function)

    def _make_docstring(self) -> str:
        """Makes a doc-string detailing the function inputs and outputs."""
        return_string = (
            f"Algorithm: {self._name}\n" + "Inputs:\n" + ", ".join(self.input_keys) + "\n" + "Outputs:\n" + ", ".join(self.return_keys)
        )
        return return_string

    def __repr__(self) -> str:
        """Return a simple string description of the Algorithm."""
        return f"Algorithm: {self._name}"

    @classmethod
    def _make_run(cls, func: FunctionType) -> Callable[..., xr.Dataset]:
        """Helper to create the `run()` function with correct doc string.

        Args:
            func: function to be wrapped

        Returns: a xarray DataSet of the result
        """

        @wraps(func)
        def run(**kwargs: Any) -> xr.Dataset:
            result = func(**kwargs)
            dataset = xr.Dataset(result)
            return dataset

        return run

    def update_dataset(self, dataset: xr.Dataset, in_place: bool = False) -> Optional[xr.Dataset]:
        """Retrieve inputs from passed dataset and return a new dataset combining input and output quantities.

        Specifying in_place=True modifies the dataset in place (changing the input), whereas in_place=False will
        return a copy of the dataset with the outputs appended.

        Args:
            dataset: input dataset
            in_place: modify the dataset in place, otherwise return a modified dataset keeping the input unchanged.

        Returns: modified dataset
        """
        if not in_place:
            dataset = dataset.copy(deep=True)

        input_values = {}
        for key in self.input_keys:
            if key in dataset.keys():
                input_values[key] = dataset[key]
            elif key in self.default_keys:
                input_values[key] = self.default_values[key]
            else:
                sorted_dataset_keys = ", ".join(sorted(dataset.keys()))  # type:ignore[arg-type]
                sorted_default_keys = ", ".join(sorted(self.default_keys))
                raise KeyError(f"Key '{key}' not in dataset keys [{sorted_dataset_keys}] or default values [{sorted_default_keys}]")

        result = self._function(**input_values)

        for key, val in result.items():
            dataset[key] = val

        if not in_place:
            return dataset
        else:
            return None

    def __add__(self, other: Union[Algorithm, CompositeAlgorithm]) -> CompositeAlgorithm:
        """Build a CompositeAlgorithm composed of this Algorithm and another Algorithm or CompositeAlgorithm."""
        if isinstance(other, CompositeAlgorithm):
            return CompositeAlgorithm(algorithms=[self, *other.algorithms])
        else:
            return CompositeAlgorithm(algorithms=[self, other])

    @classmethod
    def from_single_function(
        cls, func: Callable, return_keys: list[str], name: Optional[str] = None, skip_unit_conversion: bool = False
    ) -> Algorithm:
        """Build an Algorithm which wraps a single function."""

        @wraps(func)
        def wrapped_function(**kwargs: Any) -> dict:
            result = func(**kwargs)

            if not isinstance(result, tuple):
                result = (result,)

            result_dict = {}
            for i, key in enumerate(return_keys):
                if skip_unit_conversion:
                    result_dict[key] = result[i]
                else:
                    result_dict[key] = convert_to_default_units(result[i], key)

            return result_dict

        return cls(wrapped_function, return_keys, name=name)

    def validate_inputs(
        self, configuration: Union[dict, xr.Dataset], quiet: bool = False, raise_error_on_missing_inputs: bool = False
    ) -> bool:
        """Check that all required inputs are defined, and warn if inputs are unused."""
        return _validate_inputs(self, configuration, quiet=quiet, raise_error_on_missing_inputs=raise_error_on_missing_inputs)


class CompositeAlgorithm:
    """A class which combined multiple Algorithms into a single object which behaves like an Algorithm."""

    def __init__(self, algorithms: Sequence[Union[Algorithm, CompositeAlgorithm]], name: Optional[str] = None):
        """Initialise a CompositeAlgorithm, combining several other Algorithms.

        Args:
            algorithms: a list of Algorithms, in the order that they should be executed.
            name: a name used to refer to the composite algorithm.
        """
        if not (isinstance(algorithms, Sequence) and all(isinstance(alg, (Algorithm, CompositeAlgorithm)) for alg in algorithms)):
            raise TypeError("Should pass a list of algorithms or composites to CompositeAlgorithm.")

        self.algorithms: list[Algorithm] = []

        # flattens composite algorithms into their respective list of plain Algorithms
        for alg in algorithms:
            if isinstance(alg, Algorithm):
                self.algorithms.append(alg)
            else:
                self.algorithms.extend(alg.algorithms)

        self.input_keys: list[str] = []
        self.required_input_keys: list[str] = []
        self.return_keys: list[str] = []
        pars: list[inspect.Parameter] = []

        # traverse list of algorithms in order.
        # If an ouput from the set of previous algorithms provides an input to a following algorithm
        # the input is not turned into an input to the CompositeAlgorithm
        for alg in self.algorithms:
            alg_sig = inspect.signature(alg.run)
            for key in alg.default_keys:
                if key not in self.return_keys:
                    self.input_keys.append(key)
                    pars.append(alg_sig.parameters[key])
            for key in alg.required_input_keys:
                if key not in self.return_keys:
                    self.input_keys.append(key)
                    self.required_input_keys.append(key)
                    pars.append(alg_sig.parameters[key])

            for key in alg.return_keys:
                if key not in self.return_keys:
                    self.return_keys.append(key)

        # create a signature for the run() function
        # This is a purely aesthetic change, that ensures the run() function
        # has a helpful tooltip in editors and in the documentation

        # 1. make sure the list of pars doesn't have any duplicates, if there are duplicates
        # we pick the first one. We don't assert that the types of two parameters are compatible
        # that's not easy to do.
        seen_pars: dict[str, int] = {}
        pars = [p for i, p in enumerate(pars) if seen_pars.setdefault(p.name, i) == i]

        # ensure POSITIONAL_OR_KEYWORD are before kw only
        pars = sorted(pars, key=lambda p: p.kind)

        def_pars = [p for p in pars if p.default != inspect.Parameter.empty]
        non_def_pars = [p for p in pars if p.default == inspect.Parameter.empty]

        # methods are immutable and we don't want to set a signature on the class' run() method
        # thus we wrap the original run method and then assign the __signature__ to the wrapped
        # wrapper function
        def _wrap(f: Callable[..., xr.Dataset]) -> Callable[..., xr.Dataset]:
            def wrapper(**kwargs: Any) -> xr.Dataset:
                return f(**kwargs)

            wrapper.__doc__ = f.__doc__

            return wrapper

        self.run = _wrap(self._run)
        # ignore due to mypy bug/missing feature https://github.com/python/mypy/issues/3482
        self.run.__signature__ = inspect.Signature(  # type:ignore[attr-defined]
            non_def_pars + def_pars, return_annotation=xr.Dataset
        )
        self._name = name
        self.__doc__ = self._make_docstring()

    def _make_docstring(self) -> str:
        """Makes a doc-string detailing the function inputs and outputs."""
        components = f"[{', '.join(alg._name for alg in self.algorithms)}]"

        return_string = (
            f"CompositeAlgorithm: {self._name}\n"
            if self._name is not None
            else "CompositeAlgorithm\n"
            f"Composed of {components}\n"
            f"Inputs:\n{', '.join(self.input_keys)}\n"
            f"Outputs:\n{', '.join(self.return_keys)}"
        )
        return return_string

    def __repr__(self) -> str:
        """Return a simple string description of the CompositeAlgorithm."""
        return f"CompositeAlgorithm: {self._name}"

    def _run(self, **kwargs: Any) -> xr.Dataset:
        """Run the sub-Algorithms, one after the other and return a xarray.Dataset of the results.

        Will throw a warning if parameters are not used by any sub-Algorithm.
        """
        result = kwargs

        parameters_extra = set(kwargs) - set(self.required_input_keys)
        parameters_missing = set(self.required_input_keys) - set(kwargs)
        if parameters_missing:
            raise TypeError(f"CompositeAlgorithm.run() missing arguments: {', '.join(parameters_missing)}")
        if parameters_extra:
            warn(f"Not all input parameters were used. Unused parameters: [{', '.join(parameters_extra)}]", stacklevel=3)

        for alg in self.algorithms:

            alg_kwargs = {key: result[key] for key in result.keys() if key in alg.input_keys}

            alg_result = alg.run(**alg_kwargs)
            result.update(alg_result)  # type:ignore[arg-type]  # dict.update() doesn't like KeysView[Hashable]

        return xr.Dataset(result)

    def update_dataset(self, dataset: xr.Dataset, in_place: bool = False) -> Optional[xr.Dataset]:
        """Retrieve inputs from passed dataset and return a new dataset combining input and output quantities.

        Specifying in_place=True modifies the dataset in place (changing the input), whereas in_place=False will
        return a copy of the dataset with the outputs appended.

        N.b. will not throw a warning if the dataset contains unused elements.

        Args:
            dataset: input dataset
            in_place: modify the dataset in place, otherwise return a modified dataset keeping the input unchanged.

        Returns: modified dataset
        """
        if not in_place:
            dataset = dataset.copy(deep=True)

        for alg in self.algorithms:
            # We've already used copy on the dataset, so can now call update_dataset with
            # in_place = True for each of the algorithms.
            alg.update_dataset(dataset, in_place=True)

        if not in_place:
            return dataset
        else:
            return None

    def __add__(self, other: Union[Algorithm, CompositeAlgorithm]) -> CompositeAlgorithm:
        """Build a CompositeAlgorithm composed of this CompositeAlgorithm and another Algorithm or CompositeAlgorithm."""
        if isinstance(other, Algorithm):
            return CompositeAlgorithm(algorithms=[*self.algorithms, other])
        else:
            return CompositeAlgorithm(algorithms=[*self.algorithms, *other.algorithms])

    def validate_inputs(  # noqa: PLR0912
        self,
        configuration: Union[dict, xr.Dataset],
        quiet: bool = False,
        raise_error_on_missing_inputs: bool = True,
        warn_for_overridden_variables: bool = False,
    ) -> bool:
        """Check that all required inputs are defined, and warn if inputs are unused."""
        # Check if variables are being silently internally overwritten
        config_keys = list(configuration.keys())
        key_setter = {key: ["INPUT"] for key in config_keys}

        for algorithm in self.algorithms:
            for key in algorithm.return_keys:
                if key not in key_setter.keys():
                    key_setter[key] = [algorithm._name]
                else:
                    key_setter[key].append(algorithm._name)

        overridden_variables = []
        for variable, algs in key_setter.items():
            if len(algs) > 1:
                overridden_variables.append(f"{variable}: ({', '.join(algs)})")

        if warn_for_overridden_variables and len(overridden_variables) > 0:
            warn(
                f"The following variables were overridden internally (given as variable: (list of algorithms setting variable)): {', '.join(overridden_variables)}",
                stacklevel=3,
            )

        # Check that algorithms are ordered such that dependent algorithms follow those setting their required input keys
        available_parameters = config_keys.copy()
        out_of_order_parameters = {}
        for algorithm in self.algorithms:
            for key in algorithm.required_input_keys:
                if key not in available_parameters:
                    out_of_order_parameters[key] = algorithm
            for key in algorithm.return_keys:
                available_parameters.append(key)

        if len(out_of_order_parameters) > 0:
            message = ""
            for key, algorithm in out_of_order_parameters.items():
                if key in key_setter and len(key_setter.get(key, [])) > 0:
                    message += f"{key} needed by {algorithm} defined by output of {key_setter[key]}."
            if len(message) > 0:
                message = f"Algorithms out of order. {message}. Rearrange the list of algorithms so that dependent algorithm are after algorithms setting their inputs."
                if raise_error_on_missing_inputs:
                    raise RuntimeError(message)
                if not quiet:
                    warn(message, stacklevel=3)

            _validate_inputs(self, configuration, quiet=quiet, raise_error_on_missing_inputs=raise_error_on_missing_inputs)

            return False
        else:
            return _validate_inputs(self, configuration, quiet=quiet, raise_error_on_missing_inputs=raise_error_on_missing_inputs)


def _validate_inputs(
    algorithm: Union[Algorithm, CompositeAlgorithm],
    configuration: Union[dict, xr.Dataset],
    quiet: bool = False,
    raise_error_on_missing_inputs: bool = False,
) -> bool:
    """Check that all required inputs are defined, and warn if inputs are unused."""
    config_keys = list(configuration.keys())

    unused_config_keys = config_keys.copy()
    missing_input_keys = set(algorithm.required_input_keys)

    for key in config_keys:
        if key in missing_input_keys:
            missing_input_keys.remove(key)

        if key in algorithm.input_keys:
            # required_input_keys gives the list of keys which must
            # be provided, while input_puts gives the list of keys
            # which can be provided (but which might have default values).
            unused_config_keys.remove(key)

    if len(missing_input_keys) == 0 and len(unused_config_keys) == 0:
        return True

    elif len(missing_input_keys) > 0 and len(unused_config_keys) > 0:
        message = f"Missing input parameters [{', '.join(missing_input_keys)}]. Also had unused input parameters [{', '.join(unused_config_keys)}]."
        if raise_error_on_missing_inputs:
            raise RuntimeError(message)

    elif len(missing_input_keys) > 0:
        message = f"Missing input parameters [{', '.join(missing_input_keys)}]."
        if raise_error_on_missing_inputs:
            raise RuntimeError(message)

    else:
        message = f"Unused input parameters [{', '.join(unused_config_keys)}]."

    if not quiet:
        warn(message, stacklevel=3)
    return False
